from abc import ABC
from typing import Any

from lionagi.libs.sys_util import PATH_TYPE
from lionagi.libs import convert, dataframe, SysUtil

from ..schema.base_node import BaseRelatableNode
from ..schema.data_logger import DataLogger, DLog
from ..messages.schema import (
    BranchColumns,
    System,
    Response,
    Instruction,
    BaseMessage,
)
from .util import MessageUtil


class BaseBranch(BaseRelatableNode, ABC):
    """
    Base class for managing branches of conversation, incorporating messages
    and logging functionality.

    Attributes:
            messages (dataframe.ln_DataFrame): Holds the messages in the branch.
            datalogger (DataLogger): Logs data related to the branch's operation.
            persist_path (PATH_TYPE): Filesystem path for data persistence.
    """

    _columns: list[str] = BranchColumns.COLUMNS.value

    def __init__(
        self,
        messages: dataframe.ln_DataFrame | None = None,
        datalogger: DataLogger | None = None,
        persist_path: PATH_TYPE | None = None,
        name=None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        if isinstance(messages, dataframe.ln_DataFrame):
            if MessageUtil.validate_messages(messages):
                self.messages = messages
            else:
                raise ValueError("Invalid messages format")
        else:
            self.messages = dataframe.ln_DataFrame(columns=self._columns)

        self.datalogger = datalogger or DataLogger(persist_path=persist_path)
        self.name = name

    def add_message(
        self,
        system: dict | list | System | None = None,
        instruction: dict | list | Instruction | None = None,
        context: str | dict[str, Any] | None = None,
        response: dict | list | BaseMessage | None = None,
        output_fields=None,
        recipient=None,
        **kwargs,
    ) -> None:
        """
        Adds a message to the branch.

        Args:
                system: Information for creating a System message.
                instruction: Information for creating an Instruction message.
                context: Context information for the message.
                response: Response data for creating a message.
                **kwargs: Additional keyword arguments for message creation.
        """
        _msg = MessageUtil.create_message(
            system=system,
            instruction=instruction,
            context=context,
            response=response,
            output_fields=output_fields,
            recipient=recipient,
            **kwargs,
        )

        if isinstance(_msg, System):
            self.system_node = _msg

        # sourcery skip: merge-nested-ifs
        if isinstance(_msg, Instruction):
            if recipient is None and self.name is not None:
                _msg.recipient = self.name

        if isinstance(_msg, Response):
            if "action_response" in _msg.content.keys():
                if recipient is None and self.name is not None:
                    _msg.recipient = self.name
                if recipient is not None and self.name is None:
                    _msg.recipient = recipient
            if "response" in _msg.content.keys():
                if self.name is not None:
                    _msg.sender = self.name

        _msg.content = _msg.msg_content
        self.messages.loc[len(self.messages)] = _msg.to_pd_series()

    def _to_chatcompletion_message(
        self, with_sender: bool = False
    ) -> list[dict[str, Any]]:
        """
        Converts messages to a list of dictionaries formatted for chat completion,
        optionally including sender information.

        Args:
                with_sender: Flag to include sender information in the output.

        Returns:
                A list of message dictionaries, each with 'role' and 'content' keys,
                and optionally prefixed by 'Sender' if with_sender is True.
        """

        message = []

        for _, row in self.messages.iterrows():
            content_ = row["content"]
            if content_.startswith("Sender"):
                content_ = content_.split(":", 1)[1]

            # if isinstance(content_, str):
            #     try:
            #         content_ = json.dumps(to_dict(content_))
            #     except Exception as e:
            #         raise ValueError(
            #             f"Error in serializing, {row['node_id']} {content_}: {e}"
            #         )

            out = {"role": row["role"], "content": content_}
            if with_sender:
                out["content"] = f"Sender {row['sender']}: {content_}"

            message.append(out)
        return message

    @property
    def chat_messages(self) -> list[dict[str, Any]]:
        """
        Retrieves all chat messages without sender information.

        Returns:
                A list of dictionaries representing chat messages.
        """

        return self._to_chatcompletion_message()

    @property
    def chat_messages_with_sender(self) -> list[dict[str, Any]]:
        """
        Retrieves all chat messages, including sender information.

        Returns:
                A list of dictionaries representing chat messages, each prefixed with its sender.
        """

        return self._to_chatcompletion_message(with_sender=True)

    @property
    def last_message(self) -> dataframe.ln_DataFrame:
        """
        Retrieves the last message from the branch as a pandas Series.

        Returns:
                A pandas Series representing the last message in the branch.
        """

        return MessageUtil.get_message_rows(self.messages, n=1, from_="last")

    @property
    def last_message_content(self) -> dict[str, Any]:
        """
        Extracts the content of the last message in the branch.

        Returns:
                A dictionary representing the content of the last message.
        """

        return convert.to_dict(self.messages.content.iloc[-1])

    @property
    def first_system(self) -> dataframe.ln_DataFrame:
        """
        Retrieves the first message marked with the 'system' role.

        Returns:
                A pandas Series representing the first 'system' message in the branch.
        """

        return MessageUtil.get_message_rows(
            self.messages, role="system", n=1, from_="front"
        )

    @property
    def last_response(self) -> dataframe.ln_DataFrame:
        """
        Retrieves the last message marked with the 'assistant' role.

        Returns:
                A pandas Series representing the last 'assistant' (response) message in the branch.
        """

        return MessageUtil.get_message_rows(
            self.messages, role="assistant", n=1, from_="last"
        )

    @property
    def last_response_content(self) -> dict[str, Any]:
        """
        Extracts the content of the last 'assistant' (response) message.

        Returns:
                A dictionary representing the content of the last 'assistant' message.
        """

        return convert.to_dict(self.last_response.content.iloc[-1])

    @property
    def action_request(self) -> dataframe.ln_DataFrame:
        """
        Filters and retrieves all messages sent by 'action_request'.

        Returns:
                A pandas DataFrame containing all 'action_request' messages.
        """

        return convert.to_df(self.messages[self.messages.sender == "action_request"])

    @property
    def action_response(self) -> dataframe.ln_DataFrame:
        """
        Filters and retrieves all messages sent by 'action_response'.

        Returns:
                A pandas DataFrame containing all 'action_response' messages.
        """

        return convert.to_df(self.messages[self.messages.sender == "action_response"])

    @property
    def responses(self) -> dataframe.ln_DataFrame:
        """
        Retrieves all messages marked with the 'assistant' role.

        Returns:
                A pandas DataFrame containing all messages with an 'assistant' role.
        """

        return convert.to_df(self.messages[self.messages.role == "assistant"])

    @property
    def assistant_responses(self) -> dataframe.ln_DataFrame:
        """
        Filters 'assistant' role messages excluding 'action_request' and 'action_response'.

        Returns:
                A pandas DataFrame of 'assistant' messages excluding action requests/responses.
        """

        a_responses = self.responses[self.responses.sender != "action_response"]
        a_responses = a_responses[a_responses.sender != "action_request"]
        return convert.to_df(a_responses)

    @property
    def last_assistant_response(self):
        return self.assistant_responses.iloc[-1]

    @property
    def info(self) -> dict[str, Any]:
        """
        Summarizes branch information, including message counts by role.

        Returns:
                A dictionary containing counts of messages categorized by their role.
        """

        return self._info()

    @property
    def sender_info(self) -> dict[str, int]:
        """
        Provides a summary of message counts categorized by sender.

        Returns:
                A dictionary with senders as keys and counts of their messages as values.
        """

        return self._info(use_sender=True)

    @property
    def describe(self) -> dict[str, Any]:
        """
        Provides a detailed description of the branch, including a summary of messages.

        Returns:
                A dictionary with a summary of total messages, a breakdown by role, and
                a preview of the first five messages.
        """

        return {
            "total_messages": len(self.messages),
            "summary_by_role": self._info(),
            "messages": [msg.to_dict() for _, msg in self.messages.iterrows()][
                : len(self.messages) - 1 if len(self.messages) < 5 else 5
            ],
        }

    @classmethod
    def _from_csv(cls, filename: str, read_kwargs=None, **kwargs) -> "BaseBranch":
        read_kwargs = {} if read_kwargs is None else read_kwargs
        messages = dataframe.read_csv(filename, **read_kwargs)
        return cls(messages=messages, **kwargs)

    @classmethod
    def from_csv(cls, **kwargs) -> "BaseBranch":

        return cls._from_csv(**kwargs)

    @classmethod
    def from_json_string(cls, **kwargs) -> "BaseBranch":

        return cls._from_json(**kwargs)

    @classmethod
    def _from_json(cls, filename: str, read_kwargs=None, **kwargs) -> "BaseBranch":
        read_kwargs = {} if read_kwargs is None else read_kwargs
        messages = dataframe.read_json(filename, **read_kwargs)
        return cls(messages=messages, **kwargs)

    def to_csv_file(
        self,
        filename: PATH_TYPE = "messages.csv",
        dir_exist_ok: bool = True,
        timestamp: bool = True,
        time_prefix: bool = False,
        verbose: bool = True,
        clear: bool = True,
        **kwargs,
    ) -> None:
        """
        Exports the branch messages to a CSV file.

        Args:
                filepath: Destination path for the CSV file. Defaults to 'messages.csv'.
                dir_exist_ok: If False, an error is raised if the directory exists. Defaults to True.
                timestamp: If True, appends a timestamp to the filename. Defaults to True.
                time_prefix: If True, prefixes the filename with a timestamp. Defaults to False.
                verbose: If True, prints a message upon successful export. Defaults to True.
                clear: If True, clears the messages after exporting. Defaults to True.
                **kwargs: Additional keyword arguments for pandas.DataFrame.to_csv().
        """

        if not filename.endswith(".csv"):
            filename += ".csv"

        filename = SysUtil.create_path(
            self.datalogger.persist_path,
            filename,
            timestamp=timestamp,
            dir_exist_ok=dir_exist_ok,
            time_prefix=time_prefix,
        )

        try:
            self.messages.to_csv(filename, **kwargs)
            if verbose:
                print(f"{len(self.messages)} messages saved to {filename}")
            if clear:
                self.clear_messages()
        except Exception as e:
            raise ValueError(f"Error in saving to csv: {e}") from e

    def to_json_file(
        self,
        filename: PATH_TYPE = "messages.json",
        dir_exist_ok: bool = True,
        timestamp: bool = True,
        time_prefix: bool = False,
        verbose: bool = True,
        clear: bool = True,
        **kwargs,
    ) -> None:
        """
        Exports the branch messages to a JSON file.

        Args:
                filename: Destination path for the JSON file. Defaults to 'messages.json'.
                dir_exist_ok: If False, an error is raised if the dirctory exists. Defaults to True.
                timestamp: If True, appends a timestamp to the filename. Defaults to True.
                time_prefix: If True, prefixes the filename with a timestamp. Defaults to False.
                verbose: If True, prints a message upon successful export. Defaults to True.
                clear: If True, clears the messages after exporting. Defaults to True.
                **kwargs: Additional keyword arguments for pandas.DataFrame.to_json().
        """

        if not filename.endswith(".json"):
            filename += ".json"

        filename = SysUtil.create_path(
            self.datalogger.persist_path,
            filename,
            timestamp=timestamp,
            dir_exist_ok=dir_exist_ok,
            time_prefix=time_prefix,
        )

        try:
            self.messages.to_json(
                filename, orient="records", lines=True, date_format="iso", **kwargs
            )
            if verbose:
                print(f"{len(self.messages)} messages saved to {filename}")
            if clear:
                self.clear_messages()
        except Exception as e:
            raise ValueError(f"Error in saving to json: {e}") from e

    def log_to_csv(
        self,
        filename: PATH_TYPE = "log.csv",
        dir_exist_ok: bool = True,
        timestamp: bool = True,
        time_prefix: bool = False,
        verbose: bool = True,
        clear: bool = True,
        flatten_=True,
        sep="[^_^]",
        **kwargs,
    ) -> None:
        """
        Exports the data logger contents to a CSV file.

        Args:
                filename: Destination path for the CSV file. Defaults to 'log.csv'.
                dir_exist_ok: If False, an error is raised if the directory exists. Defaults to True.
                timestamp: If True, appends a timestamp to the filename. Defaults to True.
                time_prefix: If True, prefixes the filename with a timestamp. Defaults to False.
                verbose: If True, prints a message upon successful export. Defaults to True.
                clear: If True, clears the logger after exporting. Defaults to True.
                **kwargs: Additional keyword arguments for pandas.DataFrame.to_csv().
        """
        self.datalogger.to_csv_file(
            filename=filename,
            dir_exist_ok=dir_exist_ok,
            timestamp=timestamp,
            time_prefix=time_prefix,
            verbose=verbose,
            clear=clear,
            flatten_=flatten_,
            sep=sep,
            **kwargs,
        )

    def log_to_json(
        self,
        filename: PATH_TYPE = "log.json",
        dir_exist_ok: bool = True,
        timestamp: bool = True,
        time_prefix: bool = False,
        verbose: bool = True,
        clear: bool = True,
        flatten_=True,
        sep="[^_^]",
        **kwargs,
    ) -> None:
        """
        Exports the data logger contents to a JSON file.

        Args:
                filename: Destination path for the JSON file. Defaults to 'log.json'.
                dir_exist_ok: If False, an error is raised if the directory exists. Defaults to True.
                timestamp: If True, appends a timestamp to the filename. Defaults to True.
                time_prefix: If True, prefixes the filename with a timestamp. Defaults to False.
                verbose: If True, prints a message upon successful export. Defaults to True.
                clear: If True, clears the logger after exporting. Defaults to True.
                **kwargs: Additional keyword arguments for pandas.DataFrame.to_json().
        """

        self.datalogger.to_json_file(
            filename=filename,
            dir_exist_ok=dir_exist_ok,
            timestamp=timestamp,
            time_prefix=time_prefix,
            verbose=verbose,
            clear=clear,
            flatten_=flatten_,
            sep=sep,
            **kwargs,
        )

    def load_log(self, filename, flattened=True, sep="[^_^]", verbose=True, **kwargs):
        df = ""
        try:
            if filename.endswith(".csv"):
                df = dataframe.read_csv(filename, **kwargs)

            elif filename.endswith(".json"):
                df = dataframe.read_json(filename, **kwargs)

            for _, row in df.iterrows():
                self.datalogger.log.append(
                    DLog.deserialize(
                        input_str=row.input_data,
                        output_str=row.output_data,
                        unflatten_=flattened,
                        sep=sep,
                    )
                )

            if verbose:
                print(f"Loaded {len(df)} logs from {filename}")
        except Exception as e:
            raise ValueError(f"Error in loading log: {e}") from e

    def remove_message(self, node_id: str) -> None:
        """
        Removes a message from the branch based on its node ID.

        Args:
                node_id: The unique identifier of the message to be removed.
        """
        MessageUtil.remove_message(self.messages, node_id)

    def update_message(self, node_id: str, column: str, value: Any) -> bool:
        """
        Updates a specific column of a message identified by node_id with a new value.

        Args:
                value: The new value to update the message with.
                node_id: The unique identifier of the message to update.
                column: The column of the message to update.
        """

        index = self.messages[self.messages["node_id"] == node_id].index[0]

        return dataframe.update_row(
            self.messages, row=index, column=column, value=value
        )

    def change_first_system_message(
        self, system: str | dict[str, Any] | System, sender: str | None = None
    ) -> None:
        """
        Updates the first system message with new content and/or sender.

        Args:
                system: The new system message content or a System object.
                sender: The identifier of the sender for the system message.
        """

        if len(self.messages[self.messages["role"] == "system"]) == 0:
            raise ValueError("There is no system message in the messages.")

        if not isinstance(system, (str, dict, System)):
            raise ValueError("Input cannot be converted into a system message.")

        if isinstance(system, (str, dict)):
            system = System(system, sender=sender)

        if isinstance(system, System):
            system.timestamp = SysUtil.get_timestamp()
            sys_index = self.messages[self.messages.role == "system"].index
            self.messages.loc[sys_index[0]] = system.to_pd_series()

    def rollback(self, steps: int) -> None:
        """
        Removes the last 'n' messages from the branch.

        Args:
                steps: The number of messages to remove from the end.
        """

        self.messages = dataframe.remove_last_n_rows(self.messages, steps)

    def clear_messages(self) -> None:
        """
        Clears all messages from the branch.
        """
        self.messages = dataframe.ln_DataFrame(columns=self._columns)

    def replace_keyword(
        self,
        keyword: str,
        replacement: str,
        column: str = "content",
        case_sensitive: bool = False,
    ) -> None:

        dataframe.replace_keyword(
            self.messages,
            keyword,
            replacement,
            column=column,
            case_sensitive=case_sensitive,
        )

    def search_keywords(
        self,
        keywords: str | list[str],
        case_sensitive: bool = False,
        reset_index: bool = False,
        dropna: bool = False,
    ) -> dataframe.ln_DataFrame:
        return dataframe.search_keywords(
            self.messages,
            keywords,
            case_sensitive=case_sensitive,
            reset_index=reset_index,
            dropna=dropna,
        )

    def extend(self, messages: dataframe.ln_DataFrame, **kwargs) -> None:

        self.messages = MessageUtil.extend(self.messages, messages, **kwargs)

    def filter_by(
        self,
        role: str | None = None,
        sender: str | None = None,
        start_time=None,
        end_time=None,
        content_keywords: str | list[str] | None = None,
        case_sensitive: bool = False,
    ) -> dataframe.ln_DataFrame:

        return MessageUtil.filter_messages_by(
            self.messages,
            role=role,
            sender=sender,
            start_time=start_time,
            end_time=end_time,
            content_keywords=content_keywords,
            case_sensitive=case_sensitive,
        )

    # noinspection PyTestUnpassedFixture
    def _info(self, use_sender: bool = False) -> dict[str, int]:
        """
        Helper method to generate summaries of messages either by role or sender.

        Args:
                use_sender: If True, summary is categorized by sender. Otherwise, by role.

        Returns:
                A dictionary summarizing the count of messages either by role or sender.
        """

        messages = self.messages["sender"] if use_sender else self.messages["role"]
        result = messages.value_counts().to_dict()
        result["total"] = len(self.messages)
        return result



from typing import Tuple, Any, TypeVar, Callable

import asyncio

from lionagi.libs import func_call, convert, ParseUtil
from lionagi.core.schema import Tool, TOOL_TYPE

T = TypeVar("T", bound=Tool)


class ToolManager:
    """
    A manager class for handling the registration and invocation of tools that are subclasses of Tool.

    This class maintains a registry of tool instances, allowing for dynamic invocation based on
    tool name and provided arguments. It supports both synchronous and asynchronous tool function
    calls.

    Attributes:
            registry (dict[str, Tool]): A dictionary to hold registered tools, keyed by their names.
    """

    registry: dict = {}

    def name_existed(self, name: str) -> bool:
        """
        Checks if a tool name already exists in the registry.

        Args:
                name (str): The name of the tool to check.

        Returns:
                bool: True if the name exists, False otherwise.
        """
        return name in self.registry

    @property
    def has_tools(self):
        return self.registry != {}

    def _register_tool(self, tool: Tool) -> None:
        """
        Registers a tool in the registry. Raises a TypeError if the object is not an instance of Tool.

        Args:
                tool (Tool): The tool instance to register.

        Raises:
                TypeError: If the provided object is not an instance of Tool.
        """
        if not isinstance(tool, Tool):
            raise TypeError("Please register a Tool object.")
        name = tool.schema_["function"]["name"]
        self.registry.update({name: tool})

    async def invoke(self, func_calls: Tuple[str, dict[str, Any]]) -> Any:
        """
        Invokes a registered tool's function with the given arguments. Supports both coroutine and regular functions.

        Args:
                func_call (Tuple[str, Dict[str, Any]]): A tuple containing the function name and a dictionary of keyword arguments.

        Returns:
                Any: The result of the function call.

        Raises:
                ValueError: If the function name is not registered or if there's an error during function invocation.
        """
        name, kwargs = func_calls
        if not self.name_existed(name):
            raise ValueError(f"Function {name} is not registered.")
        tool = self.registry[name]
        func = tool.func
        parser = tool.parser
        try:
            if func_call.is_coroutine_func(func):
                tasks = [func_call.call_handler(func, **kwargs)]
                out = await asyncio.gather(*tasks)
                return parser(out[0]) if parser else out[0]
            else:
                out = func(**kwargs)
                return parser(out) if parser else out
        except Exception as e:
            raise ValueError(
                f"Error when invoking function {name} with arguments {kwargs} with error message {e}"
            ) from e

    @staticmethod
    def get_function_call(response: dict) -> Tuple[str, dict]:
        """
        Extracts a function call and arguments from a response dictionary.

        Args:
                response (dict): The response dictionary containing the function call information.

        Returns:
                Tuple[str, dict]: A tuple containing the function name and a dictionary of arguments.

        Raises:
                ValueError: If the response does not contain valid function call information.
        """
        try:
            func = response["action"][7:]
            args = convert.to_dict(response["arguments"])
            return func, args
        except Exception:
            try:
                func = response["recipient_name"].split(".")[-1]
                args = response["parameters"]
                return func, args
            except:
                raise ValueError("response is not a valid function call")

    def register_tools(self, tools: list[Tool]) -> None:
        """
        Registers multiple tools in the registry.

        Args:
                tools (list[Tool]): A list of tool instances to register.
        """
        func_call.lcall(tools, self._register_tool)

    def to_tool_schema_list(self) -> list[dict[str, Any]]:
        """
        Generates a list of schemas for all registered tools.

        Returns:
                list[dict[str, Any]]: A list of tool schemas.

        """
        return [tool.schema_ for tool in self.registry.values()]

    def parse_tool(self, tools: TOOL_TYPE, **kwargs) -> dict:
        """
        Parses tool information and generates a dictionary for tool invocation.

        Args:
                tools: Tool information which can be a single Tool instance, a list of Tool instances, a tool name, or a list of tool names.
                **kwargs: Additional keyword arguments.

        Returns:
                dict: A dictionary containing tool schema information and any additional keyword arguments.

        Raises:
                ValueError: If a tool name is provided that is not registered.
        """

        def tool_check(tool):
            if isinstance(tool, dict):
                return tool
            elif isinstance(tool, Tool):
                return tool.schema_
            elif isinstance(tool, str):
                if self.name_existed(tool):
                    tool = self.registry[tool]
                    return tool.schema_
                else:
                    raise ValueError(f"Function {tool} is not registered.")

        if isinstance(tools, bool):
            tool_kwarg = {"tools": self.to_tool_schema_list()}
            kwargs = tool_kwarg | kwargs

        else:
            if not isinstance(tools, list):
                tools = [tools]
            tool_kwarg = {"tools": func_call.lcall(tools, tool_check)}
            kwargs = tool_kwarg | kwargs

        return kwargs


def func_to_tool(
    func_: Callable | list[Callable], parser=None, docstring_style="google"
):
    """
    Transforms a given function into a Tool object, equipped with a schema derived
    from its docstring. This process involves parsing the function's docstring based
    on a specified style ('google' or 'reST') to extract relevant metadata and
    parameters, which are then used to construct a comprehensive schema for the Tool.
    This schema facilitates the integration of the function with systems or
    frameworks that rely on structured metadata for automation, documentation, or
    interface generation purposes.

    The function to be transformed can be any Callable that adheres to the
    specified docstring conventions. The resulting Tool object encapsulates the
    original function, allowing it to be utilized within environments that require
    objects with structured metadata.

    Args:
            func_ (Callable): The function to be transformed into a Tool object. This
                                              function should have a docstring that follows the
                                              specified docstring style for accurate schema generation.
            parser (Optional[Any]): An optional parser object associated with the Tool.
                                                            This parameter is currently not utilized in the
                                                            transformation process but is included for future
                                                            compatibility and extension purposes.
            docstring_style (str): The format of the docstring to be parsed, indicating
                                                       the convention used in the function's docstring.
                                                       Supports 'google' for Google-style docstrings and
                                                       'reST' for reStructuredText-style docstrings. The
                                                       chosen style affects how the docstring is parsed and
                                                       how the schema is generated.

    Returns:
            Tool: An object representing the original function wrapped as a Tool, along
                      with its generated schema. This Tool object can be used in systems that
                      require detailed metadata about functions, facilitating tasks such as
                      automatic documentation generation, user interface creation, or
                      integration with other software tools.

    Examples:
            >>> def example_function_google(param1: int, param2: str) -> bool:
            ...     '''
            ...     An example function using Google style docstrings.
            ...
            ...     Args:
            ...         param1 (int): The first parameter, demonstrating an integer input_.
            ...         param2 (str): The second parameter, demonstrating a string input_.
            ...
            ...     Returns:
            ...         bool: A boolean value, illustrating the return type.
            ...     '''
            ...     return True
            ...
            >>> tool_google = func_to_tool(example_function_google, docstring_style='google')
            >>> print(isinstance(tool_google, Tool))
            True

            >>> def example_function_reST(param1: int, param2: str) -> bool:
            ...     '''
            ...     An example function using reStructuredText (reST) style docstrings.
            ...
            ...     :param param1: The first parameter, demonstrating an integer input_.
            ...     :type param1: int
            ...     :param param2: The second parameter, demonstrating a string input_.
            ...     :type param2: str
            ...     :returns: A boolean value, illustrating the return type.
            ...     :rtype: bool
            ...     '''
            ...     return True
            ...
            >>> tool_reST = func_to_tool(example_function_reST, docstring_style='reST')
            >>> print(isinstance(tool_reST, Tool))
            True

    Note:
            The transformation process relies heavily on the accuracy and completeness of
            the function's docstring. Functions with incomplete or incorrectly formatted
            docstrings may result in incomplete or inaccurate Tool schemas.
    """

    fs = []
    funcs = convert.to_list(func_, flatten=True, dropna=True)
    parsers = convert.to_list(parser, flatten=True, dropna=True)

    if parser:
        if len(funcs) != len(parsers) != 1:
            raise ValueError(
                "Length of parser must match length of func. Except if you only pass one"
            )

        for idx in range(len(funcs)):
            f_ = lambda _f: Tool(
                func=_f,
                schema_=ParseUtil._func_to_schema(_f, style=docstring_style),
                parser=parsers[idx] if len(parsers) > 1 else parsers[0],
            )

            fs.append(f_)

    else:
        fs = func_call.lcall(
            funcs,
            lambda _f: Tool(
                func=_f, schema_=ParseUtil._func_to_schema(_f, style=docstring_style)
            ),
        )

    return fs



import aiohttp
import unittest
from unittest.mock import AsyncMock

from lionagi.libs.ln_api import *


class TestAPIUtil(unittest.TestCase):
    def test_api_method_post(self):
        session = AsyncMock(spec=aiohttp.ClientSession)
        method = APIUtil.api_method(session, "post")
        self.assertTrue(callable(method))

    def test_api_method_invalid(self):
        session = AsyncMock(spec=aiohttp.ClientSession)
        with self.assertRaises(ValueError):
            APIUtil.api_method(session, "invalid_method")


class TestAPIUtilExtended(unittest.TestCase):
    def test_api_error_with_error(self):
        response_json = {"error": "Something went wrong"}
        self.assertTrue(APIUtil.api_error(response_json))

    def test_api_error_without_error(self):
        response_json = {"result": "Success"}
        self.assertFalse(APIUtil.api_error(response_json))

    def test_api_rate_limit_error_with_rate_limit(self):
        response_json = {"error": {"message": "Rate limit exceeded"}}
        self.assertTrue(APIUtil.api_rate_limit_error(response_json))

    def test_api_rate_limit_error_without_rate_limit(self):
        response_json = {"error": {"message": "Another error"}}
        self.assertFalse(APIUtil.api_rate_limit_error(response_json))

    def test_api_endpoint_from_url_valid(self):
        valid_url = "https://api.example.com/v1/users"
        self.assertEqual(APIUtil.api_endpoint_from_url(valid_url), "users")

    def test_api_endpoint_from_url_invalid(self):
        invalid_url = "https://api.example.com/users"
        self.assertEqual(APIUtil.api_endpoint_from_url(invalid_url), "")


if __name__ == "__main__":
    unittest.main()



import unittest
import asyncio
import time
from lionagi.libs.ln_func_call import *
import unittest


class TestLCall(unittest.TestCase):
    def test_simple_function(self):
        result = lcall([1, 2, 3], lambda x: x * 2)
        self.assertEqual(result, [2, 4, 6])

    def test_function_returns_lists_flatten(self):
        result = lcall([1, 2], lambda x: [x, x], flatten=True)
        self.assertEqual(result, [1, 1, 2, 2])

    def test_function_returns_none_dropna(self):
        result = lcall([1, 2, None], lambda x: x if x else None, dropna=True)
        self.assertEqual(result, [1, 2])

    def test_function_with_kwargs(self):
        result = lcall([1, 2], lambda x, y: x + y, y=10)
        self.assertEqual(result, [11, 12])

    def test_empty_input(self):
        self.assertEqual(lcall([], lambda x: x * 2), [])

    def test_single_element(self):
        self.assertEqual(lcall(5, lambda x: x * 2), [10])

    def test_list_input_with_simple_function(self):
        self.assertEqual(lcall([1, 2, 3], lambda x: x * 2), [2, 4, 6])

    def test_flatten(self):
        self.assertEqual(
            lcall([[1, 2], [3, 4]], lambda x: x * 2, flatten=True), [2, 4, 6, 8]
        )

    def test_drop_none(self):
        self.assertEqual(lcall([1, None, 2], lambda x: x, dropna=True), [1, 2])

    def test_with_lambda_function(self):
        self.assertEqual(lcall([1, 2, 3], lambda x: x + 1), [2, 3, 4])

    def test_exception_handling(self):
        def faulty_func(x):
            raise ValueError("Faulty Function")

        with self.assertRaises(ValueError):
            lcall([1, 2, 3], faulty_func)

    def test_with_additional_arguments(self):
        self.assertEqual(lcall([1, 2, 3], lambda x, y: x + y, y=2), [3, 4, 5])


class TestAlCall(unittest.IsolatedAsyncioTestCase):

    # async def test_empty_input(self):
    #     result = await alcall([], lambda x: x * 2, flatten=True)
    #     self.assertEqual(result, [])

    async def test_single_element_with_async_function(self):
        async def async_func(x):
            return x * 2

        result = await alcall(5, async_func)
        self.assertEqual(result, [10])

    async def test_list_input_with_sync_function(self):
        result = await alcall([1, 2, 3], lambda x: x * 2)
        self.assertEqual(result, [2, 4, 6])

    async def test_flatten_with_async_function(self):
        async def async_func(x):
            return x * 2

        result = await alcall([[1, 2], [3, 4]], async_func, flatten=True)
        self.assertEqual(result, [2, 4, 6, 8])

    async def test_drop_none_with_sync_function(self):
        result = await alcall([1, None, 2], lambda x: x, dropna=True)
        self.assertEqual(result, [1, 2])

    async def test_exception_handling_in_async_function(self):
        async def faulty_func(x):
            raise ValueError("Faulty Function")

        with self.assertRaises(ValueError):
            await alcall([1, 2, 3], faulty_func)

    async def test_with_additional_arguments(self):
        async def async_func(x, y):
            return x + y

        result = await alcall([1, 2, 3], async_func, y=2)
        self.assertEqual(result, [3, 4, 5])


import asyncio


class TestTCall(unittest.IsolatedAsyncioTestCase):

    async def test_sync_function_call(self):
        def sync_func(x):
            return x * 2

        result = await tcall(sync_func, 5)
        self.assertEqual(result, 10)

    async def test_async_function_call(self):
        async def async_func(x):
            return x * 2

        result = await tcall(async_func, 5)
        self.assertEqual(result, 10)

    async def test_with_delay(self):
        async def async_func(x):
            return x * 2

        start_time = asyncio.get_event_loop().time()
        await tcall(async_func, 5, delay=1)
        elapsed = asyncio.get_event_loop().time() - start_time
        self.assertTrue(elapsed >= 1)

    async def test_error_handling(self):
        async def async_func(x):
            raise ValueError("Test Error")

        with self.assertRaises(ValueError):
            await tcall(async_func, 5, err_msg="Custom Error Message")

    async def test_execution_timing(self):
        async def async_func(x):
            return x * 2

        result, duration = await tcall(async_func, 5, timing=True)
        self.assertEqual(result, 10)
        self.assertIsInstance(duration, float)

    async def test_ignore_error(self):
        def sync_func(x):
            raise ValueError("Test Error")

        result = await tcall(sync_func, 5, ignore_err=True)
        self.assertIsNone(result)

    async def test_with_additional_arguments(self):
        async def async_func(x, y):
            return x + y

        result = await tcall(async_func, 5, y=3)
        self.assertEqual(result, 8)


class TestMCall(unittest.IsolatedAsyncioTestCase):

    async def test_single_function_single_input(self):
        async def async_func(x):
            return x * 2

        result = await mcall(5, async_func)
        self.assertEqual(result, [10])

    async def test_list_of_functions_single_input(self):
        async def async_func1(x):
            return x * 2

        async def async_func2(x):
            return x + 3

        with self.assertRaises(ValueError):
            await mcall([5], [async_func1, async_func2])

    async def test_single_function_list_of_inputs(self):
        async def async_func(x):
            return x * 2

        with self.assertRaises(ValueError):
            await mcall([1, 2, 3], async_func)

    async def test_list_of_functions_list_of_inputs_explode_false(self):
        async def async_func1(x):
            return x * 2

        async def async_func2(x):
            return x + 3

        result = await mcall([1, 2], [async_func1, async_func2])
        self.assertEqual(result, [2, 5])

    async def test_list_of_functions_list_of_inputs_explode_true(self):
        async def async_func1(x):
            return x * 2

        async def async_func2(x):
            return x + 3

        result = await mcall([1, 2], [async_func1, async_func2], explode=True)
        self.assertEqual(result, [[2, 4], [4, 5]])

    async def test_flatten_and_dropna(self):
        async def async_func1(x):
            return x * 2

        async def async_func2(x):
            return x + 3

        result = await mcall([[1, None], [None, 2]], [async_func1, async_func2])
        self.assertEqual(result, [2, 5])

    async def test_exception_handling(self):
        async def async_func(x):
            if x == 2:
                raise ValueError("Test Error")
            return x * 2

        with self.assertRaises(ValueError):
            await mcall([1, 2], [async_func, async_func])

    async def test_with_additional_arguments(self):
        async def async_func(x, y):
            return x + y

        result = await mcall([1, 2], [async_func, async_func], y=2)
        self.assertEqual(result, [3, 4])


class TestBCall(unittest.IsolatedAsyncioTestCase):

    async def test_empty_input_list(self):
        async def async_func(x):
            return x * 2

        result = await bcall([], async_func, batch_size=2)
        self.assertEqual(result, [])

    async def test_sync_function_call(self):
        def sync_func(x):
            return x * 2

        result = await bcall([1, 2, 3], sync_func, batch_size=2)
        self.assertEqual(result, [2, 4, 6])

    async def test_async_function_call(self):
        async def async_func(x):
            return x * 2

        result = await bcall([1, 2, 3], async_func, batch_size=2)
        self.assertEqual(result, [2, 4, 6])

    async def test_batch_processing(self):
        async def async_func(x):
            return x * 2

        # Test with 4 elements and a batch size of 2
        result = await bcall([1, 2, 3, 4], async_func, batch_size=2)
        self.assertEqual(result, [2, 4, 6, 8])

    async def test_exception_handling(self):
        async def async_func(x):
            if x == 2:
                raise ValueError("Test Error")
            return x * 2

        with self.assertRaises(ValueError):
            await bcall([1, 2, 3], async_func, batch_size=2)

    async def test_with_additional_arguments(self):
        async def async_func(x, y):
            return x + y

        result = await bcall([1, 2, 3], async_func, batch_size=2, y=3)
        self.assertEqual(result, [4, 5, 6])


class TestRCall(unittest.IsolatedAsyncioTestCase):

    async def test_sync_function_without_timeout(self):
        def sync_func(x):
            return x * 2

        result = await rcall(sync_func, 5)
        self.assertEqual(result, 10)

    async def test_async_function_without_timeout(self):
        async def async_func(x):
            return x * 2

        result = await rcall(async_func, 5)
        self.assertEqual(result, 10)

    async def test_timeout(self):
        async def async_func(x):
            await asyncio.sleep(2)
            return x

        with self.assertRaises(asyncio.TimeoutError):
            await rcall(async_func, 5, timeout=1)

    async def test_retry_mechanism(self):
        attempt_count = 0

        def sync_func(x):
            nonlocal attempt_count
            attempt_count += 1
            raise ValueError("Test Error")

        with self.assertRaises(ValueError):
            await rcall(sync_func, 5, retries=3)
        self.assertEqual(attempt_count, 3)  # Initial call + 3 retries

    async def test_default_value_on_exception(self):
        def sync_func(x):
            raise ValueError("Test Error")

        result = await rcall(sync_func, 5, default=10)
        self.assertEqual(result, 10)

    async def test_exception_propagation(self):
        def sync_func(x):
            raise ValueError("Test Error")

        with self.assertRaises(ValueError):
            await rcall(sync_func, 5)


class TestCallDecorator(unittest.IsolatedAsyncioTestCase):

    def test_cache_decorator_sync(self):
        @CallDecorator.cache
        def sync_func(x):
            return x * 2

        first_call = sync_func(5)
        second_call = sync_func(5)

        self.assertEqual(first_call, 10)
        self.assertIs(first_call, second_call)

    async def test_cache_decorator_async(self):
        @CallDecorator.cache
        async def async_func(x):
            return x * 2

        first_call = await async_func(5)
        second_call = await async_func(5)

        self.assertEqual(first_call, 10)
        self.assertIs(first_call, second_call)

    def test_throttling_behavior(self):
        @CallDecorator.throttle(period=2)  # 2 seconds throttle period
        def test_func():
            return time.time()

        start_time = time.time()
        first_call = test_func()
        second_call = test_func()
        end_time = time.time()

        self.assertLess(first_call - start_time, 1)  # First call should be immediate
        self.assertGreaterEqual(
            second_call - first_call, 2
        )  # Second call should be delayed
        self.assertLessEqual(
            end_time - start_time, 3
        )  # Total time should be within reasonable bounds

    def test_successive_calls_outside_throttle_period(self):
        @CallDecorator.throttle(period=1)  # 1 second throttle period
        def test_func():
            return time.time()

        first_call = test_func()
        time.sleep(1.1)  # Sleep to exceed the throttle period
        second_call = test_func()

        self.assertGreaterEqual(
            second_call - first_call, 1
        )  # Second call should not be delayed

    def test_throttling_with_different_inputs(self):
        @CallDecorator.throttle(period=2)  # 2 seconds throttle period
        def test_func(x):
            return x

        first_call = test_func(1)
        time.sleep(0.5)
        second_call = test_func(2)

        self.assertEqual(first_call, 1)
        self.assertEqual(
            second_call, 2
        )  # Should return the result of the first call due to throttling

    def test_preprocessing(self):
        def preprocess(x):
            return x * 2

        @CallDecorator.pre_post_process(preprocess, lambda x: x)
        def test_func(x):
            return x + 3

        result = test_func(2)  # Preprocess: 2 * 2 -> 4, Func: 4 + 3
        self.assertEqual(result, 7)

    def test_postprocessing(self):
        def postprocess(x):
            return x * 3

        @CallDecorator.pre_post_process(lambda x: x, postprocess)
        def test_func(x):
            return x + 2

        result = test_func(2)  # Func: 2 + 2 -> 4, Postprocess: 4 * 3
        self.assertEqual(result, 12)

    def test_preprocessing_and_postprocessing(self):
        def preprocess(x):
            return x * 2

        def postprocess(x):
            return x * 3

        @CallDecorator.pre_post_process(preprocess, postprocess)
        def test_func(x):
            return x + 1

        result = test_func(
            2
        )  # Preprocess: 2 * 2 -> 4, Func: 4 + 1 -> 5, Postprocess: 5 * 3
        self.assertEqual(result, 15)

    def test_filtering_based_on_predicate(self):
        is_even = lambda x: x % 2 == 0

        @CallDecorator.filter(is_even)
        def test_func():
            return [1, 2, 3, 4, 5]

        result = test_func()
        self.assertEqual(result, [2, 4])

    def test_no_items_satisfy_predicate(self):
        is_negative = lambda x: x < 0

        @CallDecorator.filter(is_negative)
        def test_func():
            return [1, 2, 3, 4, 5]

        result = test_func()
        self.assertEqual(result, [])

    def test_all_items_satisfy_predicate(self):
        is_positive = lambda x: x > 0

        @CallDecorator.filter(is_positive)
        def test_func():
            return [1, 2, 3, 4, 5]

        result = test_func()
        self.assertEqual(result, [1, 2, 3, 4, 5])

    def test_mapping_function(self):
        double = lambda x: x * 2

        @CallDecorator.map(double)
        def test_func():
            return [1, 2, 3, 4, 5]

        result = test_func()
        self.assertEqual(result, [2, 4, 6, 8, 10])

    def test_empty_list(self):
        double = lambda x: x * 2

        @CallDecorator.map(double)
        def test_func():
            return []

        result = test_func()
        self.assertEqual(result, [])

    def test_mapping_with_different_data_types(self):
        to_string = lambda x: str(x)

        @CallDecorator.map(to_string)
        def test_func():
            return [1, 2.5, "test", True]

        result = test_func()
        self.assertEqual(result, ["1", "2.5", "test", "True"])

    def test_reduction_function(self):
        sum_func = lambda x, y: x + y

        @CallDecorator.reduce(sum_func, 0)
        def test_func():
            return [1, 2, 3, 4, 5]

        result = test_func()
        self.assertEqual(result, 15)  # Sum of 1, 2, 3, 4, 5

    def test_empty_list(self):
        sum_func = lambda x, y: x + y

        @CallDecorator.reduce(sum_func, 0)
        def test_func():
            return []

        result = test_func()
        self.assertEqual(result, 0)  # Initial value

    def test_reduction_with_initial_value(self):
        sum_func = lambda x, y: x + y

        @CallDecorator.reduce(sum_func, 10)
        def test_func():
            return [1, 2, 3, 4, 5]

        result = test_func()
        self.assertEqual(result, 25)  # 10 (initial) + Sum of 1, 2, 3, 4, 5

    def test_function_composition(self):
        double = lambda x: x * 2
        add_five = lambda x: x + 5

        @CallDecorator.compose(add_five, double)  # First add_five, then double
        def test_func():
            return 3

        result = test_func()
        self.assertEqual(result, (3 + 5) * 2)

    def test_empty_function_list(self):
        @CallDecorator.compose()
        def test_func():
            return "test"

        result = test_func()
        self.assertEqual(result, "test")

    def test_error_handling_in_composition(self):
        def raise_error(x):
            raise ValueError("Error in function")

        @CallDecorator.compose(raise_error)
        def test_func():
            return 3

        with self.assertRaises(ValueError):
            test_func()

    async def test_concurrency_limit(self):
        current_concurrency = 0
        max_concurrent_calls = 0

        @CallDecorator.max_concurrency(limit=2)
        async def test_func():
            nonlocal current_concurrency, max_concurrent_calls
            current_concurrency += 1
            max_concurrent_calls = max(max_concurrent_calls, current_concurrency)
            await asyncio.sleep(0.1)  # Simulate async work
            current_concurrency -= 1

        await asyncio.gather(*(test_func() for _ in range(5)))
        self.assertEqual(max_concurrent_calls, 2)

    async def test_function_completing_within_limit(self):
        @CallDecorator.max_concurrency(limit=2)
        async def test_func(x):
            await asyncio.sleep(0.1)
            return x

        results = await asyncio.gather(*(test_func(i) for i in range(5)))
        self.assertEqual(results, [0, 1, 2, 3, 4])


class TestThrottleClass(unittest.TestCase):

    def test_throttling_behavior_sync(self):
        throttle_decorator = Throttle(2)  # 2 seconds throttle period

        @throttle_decorator
        def test_func():
            return time.time()

        first_call_time = test_func()
        time.sleep(1)  # Sleep less than the throttle period
        second_call_time = test_func()

        self.assertGreaterEqual(
            second_call_time - first_call_time, 2
        )  # Second call should be throttled

    def test_throttling_behavior_async(self):
        throttle_decorator = Throttle(2)  # 2 seconds throttle period

        @throttle_decorator
        async def test_func():
            return time.time()

        async def async_test():
            first_call_time = await test_func()
            await asyncio.sleep(1)  # Sleep less than the throttle period
            second_call_time = await test_func()

            self.assertGreaterEqual(
                second_call_time - first_call_time, 2
            )  # Second call should be throttled

        asyncio.run(async_test())

    def test_successive_calls_with_sufficient_delay(self):
        throttle_decorator = Throttle(1)  # 1 second throttle period

        @throttle_decorator
        def test_func():
            return time.time()

        first_call_time = test_func()
        time.sleep(1.1)  # Sleep more than the throttle period
        second_call_time = test_func()

        self.assertLess(
            second_call_time - first_call_time, 1.5
        )  # Second call should not be throttled


class TestAsyncRetryDecorator(unittest.IsolatedAsyncioTestCase):
    async def test_successful_retry(self):
        attempt = 0

        @CallDecorator.retry(retries=3, delay=1, backoff_factor=2)
        async def test_func():
            nonlocal attempt
            attempt += 1
            if attempt < 3:
                raise ValueError("Test failure")
            return "Success"

        result = await test_func()
        self.assertEqual(result, "Success")
        self.assertEqual(attempt, 3)

    async def test_retry_limit(self):
        attempt = 0

        @CallDecorator.retry(retries=2, delay=1, backoff_factor=2)
        async def test_func():
            nonlocal attempt
            attempt += 1
            raise ValueError("Test failure")

        with self.assertRaises(ValueError):
            await test_func()
        self.assertEqual(attempt, 2)

    async def test_immediate_success(self):
        attempt = 0

        @CallDecorator.retry(retries=3, delay=1, backoff_factor=2)
        async def test_func():
            nonlocal attempt
            attempt += 1
            return "Success"

        result = await test_func()
        self.assertEqual(result, "Success")
        self.assertEqual(attempt, 1)

    async def test_retry_with_delays(self):
        attempt = 0
        start_time = time.time()

        @CallDecorator.retry(retries=3, delay=1, backoff_factor=2)
        async def test_func():
            nonlocal attempt
            attempt += 1
            if attempt < 3:
                raise ValueError("Test failure")
            return "Success"

        result = await test_func()
        end_time = time.time()
        elapsed_time = end_time - start_time

        self.assertEqual(result, "Success")
        # Note: The actual delay might be slightly longer than expected due to asyncio's event loop scheduling.
        self.assertTrue(
            elapsed_time >= 3, f"Elapsed time was {elapsed_time}, expected >= 3"
        )
        self.assertEqual(attempt, 3)


if __name__ == "__main__":
    unittest.main()



from typing import Any

import pandas as pd

from lionagi.libs import ln_convert as convert

ln_DataFrame = pd.DataFrame
ln_Series = pd.Series


def extend_dataframe(
    df1: pd.DataFrame,
    df2: pd.DataFrame,
    /,
    unique_col: str = "node_id",
    keep="first",
    **kwargs,
) -> pd.DataFrame:
    """
    Merges two DataFrames while ensuring no duplicate entries based on a specified unique column.

    Args:
            df1: The primary DataFrame.
            df2: The DataFrame to merge with the primary DataFrame.
            unique_col: The column name to check for duplicate entries. Defaults to 'node_id'.
            **kwargs: Additional keyword arguments for `drop_duplicates`.

    Returns:
            A DataFrame combined from df1 and df2 with duplicates removed based on the unique column.
    """
    try:
        if len(df2.dropna(how="all")) > 0 and len(df1.dropna(how="all")) > 0:
            df = convert.to_df([df1, df2])
            df.drop_duplicates(inplace=True, subset=[unique_col], keep=keep, **kwargs)
            df_ = convert.to_df(df)
            if len(df_) > 1:
                return df_
            else:
                raise ValueError("No data to extend")

    except Exception as e:
        raise ValueError(f"Error in extending messages: {e}") from e


def search_keywords(
    df: pd.DataFrame,
    /,
    keywords: str | list[str],
    *,
    column: str = "content",
    case_sensitive: bool = False,
    reset_index: bool = False,
    dropna: bool = False,
) -> pd.DataFrame:
    """
    Filters a DataFrame for rows where a specified column contains given keywords.

    Args:
            df: The DataFrame to search through.
            keywords: A keyword or list of keywords to search for.
            col: The column to perform the search in. Defaults to "content".
            case_sensitive: Whether the search should be case-sensitive. Defaults to False.
            reset_index: Whether to reset the DataFrame's index after filtering. Defaults to False.
            dropna: Whether to drop rows with NA values before searching. Defaults to False.

    Returns:
            A filtered DataFrame containing only rows where the specified column contains
            any of the provided keywords.
    """

    if isinstance(keywords, list):
        keywords = "|".join(keywords)

    def handle_cases():
        if not case_sensitive:
            return df[df[column].str.contains(keywords, case=False)]
        else:
            return df[df[column].str.contains(keywords)]

    out = handle_cases()
    if reset_index or dropna:
        out = convert.to_df(out, reset_index=reset_index)

    return out


def replace_keyword(
    df: pd.DataFrame,
    /,
    keyword: str,
    replacement: str,
    *,
    inplace=True,
    column: str = "content",
    case_sensitive: bool = False,
) -> pd.DataFrame | bool:
    """
    Replaces occurrences of a specified keyword with a replacement string in a DataFrame column.

    Args:
            df: The DataFrame to modify.
            keyword: The keyword to be replaced.
            replacement: The string to replace the keyword with.
            col: The column in which to perform the replacement.
            case_sensitive: If True, performs a case-sensitive replacement. Defaults to False.
    """

    df_ = df.copy(deep=False) if inplace else df.copy()

    if not case_sensitive:
        df_.loc[:, column] = df_[column].str.replace(
            keyword, replacement, case=False, regex=False
        )
    else:
        df_.loc[:, column] = df_[column].str.replace(keyword, replacement, regex=False)

    return df_ if inplace else True


def read_csv(filepath: str, **kwargs) -> pd.DataFrame:
    """
    Reads a CSV file into a DataFrame with optional additional pandas read_csv parameters.

    Args:
            filepath: The path to the CSV file to read.
            **kwargs: Additional keyword arguments to pass to pandas.read_csv function.

    Returns:
            A DataFrame containing the data read from the CSV file.
    """
    df = pd.read_csv(filepath, **kwargs)
    return convert.to_df(df)


def read_json(filepath, **kwargs):
    df = pd.read_json(filepath, **kwargs)
    return convert.to_df(df)


def remove_last_n_rows(df: pd.DataFrame, steps: int) -> pd.DataFrame:
    """
    Removes the last 'n' rows from a DataFrame.

    Args:
            df: The DataFrame from which to remove rows.
            steps: The number of rows to remove from the end of the DataFrame.

    Returns:
            A DataFrame with the last 'n' rows removed.

    Raises:
            ValueError: If 'steps' is negative or greater than the number of rows in the DataFrame.
    """

    if steps < 0 or steps > len(df):
        raise ValueError(
            "'steps' must be a non-negative integer less than or equal to "
            "the length of DataFrame."
        )
    return convert.to_df(df[:-steps])


def update_row(df: pd.DataFrame, row: str | int, column: str | int, value: Any) -> bool:
    """
    Updates a row's value for a specified column in a DataFrame.

    Args:
            df: The DataFrame to update.
            col: The column whose value is to be updated.
            old_value: The current value to search for in the specified column.
            new_value: The new value to replace the old value with.

    Returns:
            True if the update was successful, False otherwise.
    """

    try:
        df.loc[row, column] = value
        return True
    except Exception:
        return False



from collections.abc import Sequence, Mapping

from abc import ABC
from dataclasses import dataclass

import logging
import re
import asyncio
import aiohttp

from typing import Any, NoReturn, Type, Callable

from lionagi.libs.ln_async import AsyncUtil
import lionagi.libs.ln_convert as convert
import lionagi.libs.ln_nested as nested
import lionagi.libs.ln_func_call as func_call


class APIUtil:
    """
    A utility class for assisting with common API usage patterns.
    """

    @staticmethod
    def api_method(
        http_session: aiohttp.ClientSession, method: str = "post"
    ) -> Callable:
        """
        Returns the corresponding HTTP method function from the http_session object.

        Args:
                http_session: The session object from the aiohttp library.
                method: The HTTP method as a string.

        Returns:
                The Callable for the specified HTTP method.

        Raises:
                ValueError: If the method is not one of the allowed ones.

        Examples:
                >>> session = aiohttp.ClientSession()
                >>> post_method = APIUtil.api_method(session, "post")
                >>> print(post_method)
                <bound method ClientSession._request of <aiohttp.client.ClientSession object at 0x...>>
        """
        if method in {"post", "delete", "head", "options", "patch"}:
            return getattr(http_session, method)
        else:
            raise ValueError(
                "Invalid request, method must be in ['post', 'delete', 'head', 'options', 'patch']"
            )

    @staticmethod
    def api_error(response_json: Mapping[str, Any]) -> bool:
        """
        Checks if the given response_json dictionary contains an "error" key.

        Args:
                response_json: The JSON assistant_response as a dictionary.

        Returns:
                True if there is an error, False otherwise.

        Examples:
                >>> response_json_with_error = {"error": "Something went wrong"}
                >>> APIUtil.api_error(response_json_with_error)
                True
                >>> response_json_without_error = {"result": "Success"}
                >>> APIUtil.api_error(response_json_without_error)
                False
        """
        if "error" in response_json:
            logging.warning(f"API call failed with error: {response_json['error']}")
            return True
        return False

    @staticmethod
    def api_rate_limit_error(response_json: Mapping[str, Any]) -> bool:
        """
        Checks if the error message in the response_json dictionary contains the phrase "Rate limit".

        Args:
                response_json: The JSON assistant_response as a dictionary.

        Returns:
                True if the phrase "Rate limit" is found, False otherwise.

        Examples:
                >>> response_json_with_rate_limit = {"error": {"message": "Rate limit exceeded"}}
                >>> api_rate_limit_error(response_json_with_rate_limit)
                True
                >>> response_json_without_rate_limit = {"error": {"message": "Another error"}}
                >>> api_rate_limit_error(response_json_without_rate_limit)
                False
        """
        return "Rate limit" in response_json.get("error", {}).get("message", "")

    @staticmethod
    @func_call.lru_cache(maxsize=128)
    def api_endpoint_from_url(request_url: str) -> str:
        """
        Extracts the API endpoint from a given URL using a regular expression.

        Args:
                request_url: The full URL to the API endpoint.

        Returns:
                The extracted endpoint or an empty string if the pattern does not match.

        Examples:
                >>> valid_url = "https://api.example.com/v1/users"
                >>> api_endpoint_from_url(valid_url)
                'users'
                >>> invalid_url = "https://api.example.com/users"
                >>> api_endpoint_from_url(invalid_url)
                ''
        """
        match = re.search(r"^https://[^/]+(/.+)?/v\d+/(.+)$", request_url)
        return match[2] if match else ""

    @staticmethod
    async def unified_api_call(
        http_session: aiohttp.ClientSession, method: str, url: str, **kwargs
    ) -> Any:
        """
        Makes an API call and automatically retries on rate limit error.

        Args:
                http_session: The session object from the aiohttp library.
                method: The HTTP method as a string.
                url: The URL to which the request is made.
                **kwargs: Additional keyword arguments to pass to the API call.

        Returns:
                The JSON assistant_response as a dictionary.

        Examples:
                >>> session = aiohttp.ClientSession()
                >>> success_url = "https://api.example.com/v1/success"
                >>> print(await unified_api_call(session, 'get', success_url))
                {'result': 'Success'}
                >>> rate_limit_url = "https://api.example.com/v1/rate_limit"
                >>> print(await unified_api_call(session, 'get', rate_limit_url))
                {'error': {'message': 'Rate limit exceeded'}}
        """
        api_call = APIUtil.api_method(http_session, method)
        retry_count = 3
        retry_delay = 5  # seconds

        for attempt in range(retry_count):
            async with api_call(url, **kwargs) as response:
                response_json = await response.json()

                if not APIUtil.api_error(response_json):
                    return response_json

                if (
                    APIUtil.api_rate_limit_error(response_json)
                    and attempt < retry_count - 1
                ):
                    logging.warning(
                        f"Rate limit error detected. Retrying in {retry_delay} seconds..."
                    )
                    await AsyncUtil.sleep(retry_delay)
                else:
                    break

        return response_json

    @staticmethod
    def get_cache_key(url: str, params: Mapping[str, Any] | None) -> str:
        """
        Creates a unique cache key based on the URL and parameters.
        """
        import hashlib

        param_str = convert.to_str(params, sort_keys=True) if params else ""
        return hashlib.md5((url + param_str).encode("utf-8")).hexdigest()

    @staticmethod
    async def retry_api_call(
        http_session: aiohttp.ClientSession,
        url: str,
        retries: int = 3,
        backoff_factor: float = 0.5,
        **kwargs,
    ) -> Any:
        """
        Retries an API call on failure, with exponential backoff.

        Args:
                http_session: The aiohttp client session.
                url: The URL to make the API call.
                retries: The number of times to retry.
                backoff_factor: The backoff factor for retries.
                **kwargs: Additional arguments for the API call.

        Returns:
                The assistant_response from the API call, if successful; otherwise, None.
        """
        for attempt in range(retries):
            try:
                async with http_session.get(url, **kwargs) as response:
                    response.raise_for_status()
                    return await response.json()
            except aiohttp.ClientError:
                if attempt < retries - 1:
                    delay = backoff_factor * (2**attempt)
                    logging.info(f"Retrying {url} in {delay} seconds...")
                    await AsyncUtil.sleep(delay)
                else:
                    logging.error(
                        f"Failed to retrieve data from {url} after {retries} attempts."
                    )
                    return None

    @staticmethod
    async def upload_file_with_retry(
        http_session: aiohttp.ClientSession,
        url: str,
        file_path: str,
        param_name: str = "file",
        additional_data: Mapping[str, Any] = None,
        retries: int = 3,
    ) -> Any:
        """
        Uploads a file to a specified URL with a retry mechanism for handling failures.

        Args:
                http_session: The HTTP session object to use for making the request.
                url: The URL to which the file will be uploaded.
                file_path: The path to the file that will be uploaded.
                param_name: The name of the parameter expected by the server for the file upload.
                additional_data: Additional data to be sent with the upload.
                retries: The number of times to retry the upload in case of failure.

        Returns:
                The HTTP assistant_response object.

        Examples:
                >>> session = aiohttp.ClientSession()
                >>> assistant_response = await APIUtil.upload_file_with_retry(session, 'http://example.com/upload', 'path/to/file.txt')
                >>> assistant_response.status
                200
        """
        for attempt in range(retries):
            try:
                with open(file_path, "rb") as file:
                    files = {param_name: file}
                    additional_data = additional_data or {}
                    async with http_session.post(
                        url, data={**files, **additional_data}
                    ) as response:
                        response.raise_for_status()
                        return await response.json()
            except aiohttp.ClientError as e:
                if attempt == retries - 1:
                    raise e
                backoff = 2**attempt
                logging.info(f"Retrying {url} in {backoff} seconds...")
                await AsyncUtil.sleep(backoff)

    @staticmethod
    @AsyncUtil.cached(ttl=10 * 60)  # Cache the result for 10 minutes
    async def get_oauth_token_with_cache(
        http_session: aiohttp.ClientSession,
        auth_url: str,
        client_id: str,
        client_secret: str,
        scope: str,
    ) -> str:
        """
        Retrieves an OAuth token from the authentication server and caches it to avoid unnecessary requests.

        Args:
                http_session: The HTTP session object to use for making the request.
                auth_url: The URL of the authentication server.
                client_id: The client ID for OAuth authentication.
                client_secret: The client secret for OAuth authentication.
                scope: The scope for which the OAuth token is requested.

        Returns:
                The OAuth token as a string.

        Examples:
                >>> session = aiohttp.ClientSession()
                >>> token = await APIUtil.get_oauth_token_with_cache(session, 'http://auth.example.com', 'client_id', 'client_secret', 'read')
                >>> token
                'mock_access_token'
        """
        async with http_session.post(
            auth_url,
            data={
                "grant_type": "client_credentials",
                "client_id": client_id,
                "client_secret": client_secret,
                "scope": scope,
            },
        ) as auth_response:
            auth_response.raise_for_status()
            return (await auth_response.json()).get("access_token")

    @staticmethod
    @AsyncUtil.cached(ttl=10 * 60)
    async def cached_api_call(
        http_session: aiohttp.ClientSession, url: str, **kwargs
    ) -> Any:
        """
        Makes an API call.

        Args:
                http_session: The aiohttp client session.
                url: The URL for the API call.
                **kwargs: Additional arguments for the API call.

        Returns:
                The assistant_response from the API call, if successful; otherwise, None.
        """
        try:
            async with http_session.get(url, **kwargs) as response:
                response.raise_for_status()
                return await response.json()
        except aiohttp.ClientError as e:
            logging.error(f"API call to {url} failed: {e}")
            return None

    @staticmethod
    def calculate_num_token(
        payload: Mapping[str, Any] = None,
        api_endpoint: str = None,
        token_encoding_name: str = None,
    ) -> int:  # sourcery skip: avoid-builtin-shadow
        """
        Calculates the number of tokens required for a request based on the payload and API endpoint.

        The token calculation logic might vary based on different API endpoints and payload content.
        This method should be implemented in a subclass to provide the specific calculation logic
        for the OpenAI API.

        Parameters:
                payload (Mapping[str, Any]): The payload of the request.

                api_endpoint (str): The specific API endpoint for the request.

                token_encoding_name (str): The name of the token encoding method.

        Returns:
                int: The estimated number of tokens required for the request.

        Example:
                >>> rate_limiter = OpenAIRateLimiter(100, 200)
                >>> payload = {'prompt': 'Translate the following text:', 'max_tokens': 50}
                >>> rate_limiter.calculate_num_token(payload, 'completions')
                # Expected token calculation for the given payload and endpoint.
        """
        import tiktoken

        encoding = tiktoken.get_encoding(token_encoding_name)
        if api_endpoint.endswith("completions"):
            max_tokens = payload.get("max_tokens", 15)
            n = payload.get("n", 1)
            completion_tokens = n * max_tokens

            # chat completions
            if api_endpoint.startswith("chat/"):
                num_tokens = 0
                for message in payload["messages"]:
                    num_tokens += 4  # every message follows <im_start>{role/name}\n{content}<im_end>\n
                    for key, value in message.items():
                        num_tokens += len(encoding.encode(value))
                        if key == "name":  # if there's a name, the role is omitted
                            num_tokens -= (
                                1
                                # role is always required and always 1 token
                            )
                num_tokens += 2  # every reply is primed with <im_start>assistant
                return num_tokens + completion_tokens
            else:
                prompt = payload["prompt"]
                if isinstance(prompt, str):  # single prompt
                    prompt_tokens = len(encoding.encode(prompt))
                    return prompt_tokens + completion_tokens
                elif isinstance(prompt, list):  # multiple prompts
                    prompt_tokens = sum(len(encoding.encode(p)) for p in prompt)
                    return prompt_tokens + completion_tokens * len(prompt)
                else:
                    raise TypeError(
                        'Expecting either string or list of strings for "prompt" field in completion request'
                    )
        elif api_endpoint == "embeddings":
            input = payload["input"]
            if isinstance(input, str):  # single input
                return len(encoding.encode(input))
            elif isinstance(input, list):  # multiple inputs
                return sum(len(encoding.encode(i)) for i in input)
            else:
                raise TypeError(
                    'Expecting either string or list of strings for "inputs" field in embedding request'
                )
        else:
            raise NotImplementedError(
                f'API endpoint "{api_endpoint}" not implemented in this script'
            )

    @staticmethod
    def create_payload(input_, config, required_, optional_, input_key, **kwargs):
        config = {**config, **kwargs}
        payload = {input_key: input_}

        for key in required_:
            payload[key] = config[key]

        for key in optional_:
            if bool(config[key]) and convert.strip_lower(config[key]) != "none":
                payload[key] = config[key]

        return payload


@dataclass
class StatusTracker:
    """
    Keeps track of various task statuses within a system.

    Attributes:
            num_tasks_started (int): The number of tasks that have been initiated.
            num_tasks_in_progress (int): The number of tasks currently being processed.
            num_tasks_succeeded (int): The number of tasks that have completed successfully.
            num_tasks_failed (int): The number of tasks that have failed.
            num_rate_limit_errors (int): The number of tasks that failed due to rate limiting.
            num_api_errors (int): The number of tasks that failed due to API errors.
            num_other_errors (int): The number of tasks that failed due to other errors.

    Examples:
            >>> tracker = StatusTracker()
            >>> tracker.num_tasks_started += 1
            >>> tracker.num_tasks_succeeded += 1
    """

    num_tasks_started: int = 0
    num_tasks_in_progress: int = 0
    num_tasks_succeeded: int = 0
    num_tasks_failed: int = 0
    num_rate_limit_errors: int = 0
    num_api_errors: int = 0
    num_other_errors: int = 0


class BaseRateLimiter(ABC):
    """
    Abstract base class for implementing rate limiters.

    This class provides the basic structure for rate limiters, including
    the replenishment of request and token capacities at regular intervals.

    Attributes:
            interval: The time interval in seconds for replenishing capacities.
            max_requests: The maximum number of requests allowed per interval.
            max_tokens: The maximum number of tokens allowed per interval.
            available_request_capacity: The current available request capacity.
            available_token_capacity: The current available token capacity.
            rate_limit_replenisher_task: The asyncio task for replenishing capacities.
    """

    def __init__(
        self,
        max_requests: int,
        max_tokens: int,
        interval: int = 60,
        token_encoding_name=None,
    ) -> None:
        self.interval: int = interval
        self.max_requests: int = max_requests
        self.max_tokens: int = max_tokens
        self.available_request_capacity: int = max_requests
        self.available_token_capacity: int = max_tokens
        self.rate_limit_replenisher_task: AsyncUtil.Task | None = None
        self._stop_replenishing: AsyncUtil.Event = AsyncUtil.create_event()
        self._lock: AsyncUtil.Lock = AsyncUtil.create_lock()
        self.token_encoding_name = token_encoding_name

    async def start_replenishing(self) -> NoReturn:
        """Starts the replenishment of rate limit capacities at regular intervals."""
        try:
            while not self._stop_replenishing.is_set():
                await AsyncUtil.sleep(self.interval)
                async with self._lock:
                    self.available_request_capacity = self.max_requests
                    self.available_token_capacity = self.max_tokens
        except asyncio.CancelledError:
            logging.info("Rate limit replenisher task cancelled.")
        except Exception as e:
            logging.error(f"An error occurred in the rate limit replenisher: {e}")

    async def stop_replenishing(self) -> None:
        """Stops the replenishment task."""
        if self.rate_limit_replenisher_task:
            self.rate_limit_replenisher_task.cancel()
            await self.rate_limit_replenisher_task
        self._stop_replenishing.set()

    async def request_permission(self, required_tokens) -> bool:
        """Requests permission to make an API call.

        Returns True if the request can be made immediately, otherwise False.
        """
        async with self._lock:
            if (
                self.available_request_capacity > 0
                and self.available_token_capacity > 0
            ):
                self.available_request_capacity -= 1
                self.available_token_capacity -= (
                    required_tokens
                    # Assuming 1 token per request for simplicity
                )
                return True
            return False

    async def _call_api(
        self,
        http_session,
        endpoint: str,
        base_url: str,
        api_key: str,
        max_attempts: int = 3,
        method: str = "post",
        payload: Mapping[str, any] = None,
        **kwargs,
    ) -> Mapping[str, any] | None:
        """
        Makes an API call to the specified endpoint using the provided HTTP session.

        Args:
                http_session: The aiohttp client session to use for the API call.
                endpoint: The API endpoint to call.
                base_url: The base URL of the API.
                api_key: The API key for authentication.
                max_attempts: The maximum number of attempts for the API call.
                method: The HTTP method to use for the API call.
                payload: The payload to send with the API call.

        Returns:
                The JSON assistant_response from the API call if successful, otherwise None.
        """
        endpoint = APIUtil.api_endpoint_from_url(base_url + endpoint)
        while True:
            if (
                self.available_request_capacity < 1
                or self.available_token_capacity < 10
            ):  # Minimum token count
                await AsyncUtil.sleep(1)  # Wait for capacity
                continue
            required_tokens = APIUtil.calculate_num_token(
                payload, endpoint, self.token_encoding_name, **kwargs
            )

            if await self.request_permission(required_tokens):
                request_headers = {"Authorization": f"Bearer {api_key}"}
                attempts_left = max_attempts

                while attempts_left > 0:
                    try:
                        method = APIUtil.api_method(http_session, method)
                        async with method(
                            url=(base_url + endpoint),
                            headers=request_headers,
                            json=payload,
                        ) as response:
                            response_json = await response.json()

                            if "error" not in response_json:
                                return response_json
                            logging.warning(
                                f"API call failed with error: {response_json['error']}"
                            )
                            attempts_left -= 1

                            if "Rate limit" in response_json["error"].get(
                                "message", ""
                            ):
                                await AsyncUtil.sleep(15)
                    except Exception as e:
                        logging.warning(f"API call failed with exception: {e}")
                        attempts_left -= 1

                logging.error("API call failed after all attempts.")
                break
            else:
                await AsyncUtil.sleep(1)

    @classmethod
    async def create(
        cls,
        max_requests: int,
        max_tokens: int,
        interval: int = 60,
        token_encoding_name=None,
    ) -> "BaseRateLimiter":
        """
        Creates an instance of BaseRateLimiter and starts the replenisher task.

        Args:
                max_requests: The maximum number of requests allowed per interval.
                max_tokens: The maximum number of tokens allowed per interval.
                interval: The time interval in seconds for replenishing capacities.
                token_encoding_name: The name of the token encoding to use.

        Returns:
                An instance of BaseRateLimiter with the replenisher task started.
        """
        instance = cls(max_requests, max_tokens, interval, token_encoding_name)
        instance.rate_limit_replenisher_task = AsyncUtil.create_task(
            instance.start_replenishing(), obj=False
        )
        return instance


class SimpleRateLimiter(BaseRateLimiter):
    """
    A simple implementation of a rate limiter.

    Inherits from BaseRateLimiter and provides a basic rate limiting mechanism.
    """

    def __init__(
        self,
        max_requests: int,
        max_tokens: int,
        interval: int = 60,
        token_encoding_name=None,
    ) -> None:
        """Initializes the SimpleRateLimiter with the specified parameters."""
        super().__init__(max_requests, max_tokens, interval, token_encoding_name)


class EndPoint:
    """
    Represents an API endpoint with rate limiting capabilities.

    This class encapsulates the details of an API endpoint, including its rate limiter.

    Attributes:
            endpoint (str): The API endpoint path.
            rate_limiter_class (Type[li.BaseRateLimiter]): The class used for rate limiting requests to the endpoint.
            max_requests (int): The maximum number of requests allowed per interval.
            max_tokens (int): The maximum number of tokens allowed per interval.
            interval (int): The time interval in seconds for replenishing rate limit capacities.
            config (Mapping): Configuration parameters for the endpoint.
            rate_limiter (Optional[li.BaseRateLimiter]): The rate limiter instance for this endpoint.

    Examples:
            # Example usage of EndPoint with SimpleRateLimiter
            endpoint = EndPoint(
                    max_requests=100,
                    max_tokens=1000,
                    interval=60,
                    endpoint_='chat/completions',
                    rate_limiter_class=li.SimpleRateLimiter,
                    config={'param1': 'value1'}
            )
            asyncio.run(endpoint.init_rate_limiter())
    """

    def __init__(
        self,
        max_requests: int = 1000,
        max_tokens: int = 100000,
        interval: int = 60,
        endpoint_: str | None = None,
        rate_limiter_class: Type[BaseRateLimiter] = SimpleRateLimiter,
        token_encoding_name=None,
        config: Mapping = None,
    ) -> None:
        self.endpoint = endpoint_ or "chat/completions"
        self.rate_limiter_class = rate_limiter_class
        self.max_requests = max_requests
        self.max_tokens = max_tokens
        self.interval = interval
        self.token_encoding_name = token_encoding_name
        self.config = config or {}
        self.rate_limiter: BaseRateLimiter | None = None
        self._has_initialized = False

    async def init_rate_limiter(self) -> None:
        """Initializes the rate limiter for the endpoint."""
        self.rate_limiter = await self.rate_limiter_class.create(
            self.max_requests, self.max_tokens, self.interval, self.token_encoding_name
        )
        self._has_initialized = True


class BaseService:
    """
    Base class for services that interact with API endpoints.

    This class provides a foundation for services that need to make API calls with rate limiting.

    Attributes:
            api_key (Optional[str]): The API key used for authentication.
            schema (Mapping[str, Any]): The schema defining the service's endpoints.
            status_tracker (StatusTracker): The object tracking the status of API calls.
            endpoints (Mapping[str, EndPoint]): A dictionary of endpoint objects.
    """

    base_url: str = ""
    available_endpoints: list = []

    def __init__(
        self,
        api_key: str | None = None,
        schema: Mapping[str, Any] = None,
        token_encoding_name: str = None,
        max_tokens: int = 100_000,
        max_requests: int = 1_000,
        interval: int = 60,
    ) -> None:
        self.api_key = api_key
        self.schema = schema or {}
        self.status_tracker = StatusTracker()
        self.endpoints: Mapping[str, EndPoint] = {}
        self.token_encoding_name = token_encoding_name
        self.chat_config_rate_limit = {
            "max_requests": max_requests,
            "max_tokens": max_tokens,
            "interval": interval,
        }

    async def init_endpoint(
        self,
        endpoint_: Sequence | str | EndPoint | None = None,
    ) -> None:
        """
        Initializes the specified endpoint or all endpoints if none is specified.

        Args:
                endpoint_: The endpoint(s) to initialize. Can be a string, an EndPoint, a list of strings, or a list of EndPoints.
        """

        if endpoint_:
            if not isinstance(endpoint_, list):
                endpoint_ = [endpoint_]
            for ep in endpoint_:
                if ep not in self.available_endpoints:
                    raise ValueError(
                        f"Endpoint {ep} not available for service {self.__class__.__name__}"
                    )

                if ep not in self.endpoints:
                    endpoint_config = nested.nget(self.schema, [ep, "config"])
                    self.schema.get(ep, {})
                    if isinstance(ep, EndPoint):
                        self.endpoints[ep.endpoint] = ep
                    elif ep == "chat/completions":
                        self.endpoints[ep] = EndPoint(
                            max_requests=self.chat_config_rate_limit.get(
                                "max_requests", 1000
                            ),
                            max_tokens=self.chat_config_rate_limit.get(
                                "max_tokens", 100000
                            ),
                            interval=self.chat_config_rate_limit.get("interval", 60),
                            endpoint_=ep,
                            token_encoding_name=self.token_encoding_name,
                            config=endpoint_config,
                        )
                    else:
                        self.endpoints[ep] = EndPoint(
                            max_requests=(
                                endpoint_config.get("max_requests", 1000)
                                if endpoint_config.get("max_requests", 1000) is not None
                                else 1000
                            ),
                            max_tokens=(
                                endpoint_config.get("max_tokens", 100000)
                                if endpoint_config.get("max_tokens", 100000) is not None
                                else 100000
                            ),
                            interval=(
                                endpoint_config.get("interval", 60)
                                if endpoint_config.get("interval", 60) is not None
                                else 60
                            ),
                            endpoint_=ep,
                            token_encoding_name=self.token_encoding_name,
                            config=endpoint_config,
                        )

                if not self.endpoints[ep]._has_initialized:
                    await self.endpoints[ep].init_rate_limiter()

        else:
            for ep in self.available_endpoints:
                endpoint_config = nested.nget(self.schema, [ep, "config"])
                self.schema.get(ep, {})
                if ep not in self.endpoints:
                    self.endpoints[ep] = EndPoint(
                        max_requests=endpoint_config.get("max_requests", 1000),
                        max_tokens=endpoint_config.get("max_tokens", 100000),
                        interval=endpoint_config.get("interval", 60),
                        endpoint_=ep,
                        token_encoding_name=self.token_encoding_name,
                        config=endpoint_config,
                    )
                if not self.endpoints[ep]._has_initialized:
                    await self.endpoints[ep].init_rate_limiter()

    async def call_api(self, payload, endpoint, method, **kwargs):
        """
        Calls the specified API endpoint with the given payload and method.

        Args:
                payload: The payload to send with the API call.
                endpoint: The endpoint to call.
                method: The HTTP method to use for the call.

        Returns:
                The assistant_response from the API call.

        Raises:
                ValueError: If the endpoint has not been initialized.
        """
        if endpoint not in self.endpoints.keys():
            raise ValueError(f"The endpoint {endpoint} has not initialized.")
        async with aiohttp.ClientSession() as http_session:
            return await self.endpoints[endpoint].rate_limiter._call_api(
                http_session=http_session,
                endpoint=endpoint,
                base_url=self.base_url,
                api_key=self.api_key,
                method=method,
                payload=payload,
                **kwargs,
            )


class PayloadPackage:

    @classmethod
    def chat_completion(cls, messages, llmconfig, schema, **kwargs):
        """
        Creates a payload for the chat completion operation.

        Args:
                messages: The messages to include in the chat completion.
                llmconfig: Configuration for the language model.
                schema: The schema describing required and optional fields.
                **kwargs: Additional keyword arguments.

        Returns:
                The constructed payload.
        """
        return APIUtil.create_payload(
            input_=messages,
            config=llmconfig,
            required_=schema["required"],
            optional_=schema["optional"],
            input_key="messages",
            **kwargs,
        )

    @classmethod
    def fine_tuning(cls, training_file, llmconfig, schema, **kwargs):
        """
        Creates a payload for the fine-tuning operation.

        Args:
                training_file: The file containing training data.
                llmconfig: Configuration for the language model.
                schema: The schema describing required and optional fields.
                **kwargs: Additional keyword arguments.

        Returns:
                The constructed payload.
        """
        return APIUtil._create_payload(
            input_=training_file,
            config=llmconfig,
            required_=schema["required"],
            optional_=schema["optional"],
            input_key="training_file",
            **kwargs,
        )



from pydantic import BaseModel, Field, ValidationError, AliasChoices, field_serializer

ln_BaseModel = BaseModel
ln_Field = Field
ln_field_serializer = field_serializer
ln_AliasChoices = AliasChoices
ln_ValidationError = ValidationError



from os import getenv
from lionagi.integrations.config.openrouter_configs import openrouter_schema
from lionagi.libs.ln_api import BaseService, PayloadPackage


class OpenRouterService(BaseService):
    base_url = "https://openrouter.ai/api/v1/"
    available_endpoints = ["chat/completions"]
    schema = openrouter_schema
    key_scheme = "OPENROUTER_API_KEY"
    token_encoding_name = "cl100k_base"

    def __init__(
        self,
        api_key=None,
        key_scheme=None,
        schema=None,
        token_encoding_name: str = "cl100k_base",
        **kwargs,
    ):
        key_scheme = key_scheme or self.key_scheme
        super().__init__(
            api_key=api_key or getenv(key_scheme),
            schema=schema or self.schema,
            token_encoding_name=token_encoding_name,
            **kwargs,
        )
        self.active_endpoint = []

    async def serve(self, input_, endpoint="chat/completions", method="post", **kwargs):
        if endpoint not in self.active_endpoint:
            await self.init_endpoint(endpoint)
        if endpoint == "chat/completions":
            return await self.serve_chat(input_, **kwargs)
        else:
            return ValueError(f"{endpoint} is currently not supported")

    async def serve_chat(self, messages, **kwargs):
        endpoint = "chat/completions"

        if endpoint not in self.active_endpoint:
            await self.init_endpoint(endpoint)
            self.active_endpoint.append(endpoint)
        payload = PayloadPackage.chat_completion(
            messages, self.endpoints[endpoint].config, self.schema[endpoint], **kwargs
        )

        try:
            completion = await self.call_api(payload, endpoint, "post")
            return payload, completion
        except Exception as e:
            self.status_tracker.num_tasks_failed += 1
            raise e



import unittest
from datetime import datetime, timezone
from unittest.mock import patch, MagicMock
import shutil
import tempfile
import time
import sys
from pathlib import Path

from lionagi.libs.sys_util import SysUtil


class TestSysUtil(unittest.TestCase):

    def test_sleep(self):
        start = time.time()
        SysUtil.sleep(0.1)  # Sleep for 0.1 seconds
        end = time.time()
        self.assertTrue((end - start) >= 0.1)

    def test_get_now(self):
        # Test returning Unix timestamp
        timestamp = time.time()
        self.assertIsInstance(timestamp, float)

        # Test returning datetime object
        now = SysUtil.get_now(datetime_=True)
        self.assertIsInstance(now, datetime)

    def test_change_dict_key(self):
        dict_ = {"old_key": "value"}
        SysUtil.change_dict_key(dict_, "old_key", "new_key")
        self.assertIn("new_key", dict_)
        self.assertNotIn("old_key", dict_)

    def test_get_timestamp(self):
        timestamp = SysUtil.get_timestamp()
        self.assertIsInstance(timestamp, str)
        # Test custom separator
        custom_sep_timestamp = SysUtil.get_timestamp(sep="-")
        self.assertIn("-", custom_sep_timestamp)

    def test_create_copy(self):
        input_ = {"key": "value"}
        # Single copy
        single_copy = SysUtil.create_copy(input_)
        self.assertEqual(input_, single_copy)
        self.assertNotEqual(id(input_), id(single_copy))
        # Multiple copies
        copies = SysUtil.create_copy(input_, 2)
        self.assertIsInstance(copies, list)
        self.assertEqual(len(copies), 2)
        self.assertNotEqual(id(copies[0]), id(copies[1]))

    def test_create_id(self):
        id1 = SysUtil.create_id()
        id2 = SysUtil.create_id()
        self.assertIsInstance(id1, str)
        self.assertEqual(len(id1), 32)
        self.assertNotEqual(id1, id2)

    def test_get_bins(self):
        input_ = ["a" * 500, "b" * 1000, "c" * 500, "d" * 1000]
        bins = SysUtil.get_bins(input_)
        self.assertEqual(len(bins), 2)
        self.assertEqual(bins, [[0, 1], [2, 3]])

    def test_get_cpu_architecture(self):
        architecture = SysUtil.get_cpu_architecture()
        self.assertIn(architecture, ["apple_silicon", "other_cpu"])

    def test_is_package_installed(self):
        with patch("importlib.util.find_spec", return_value=None):
            self.assertFalse(SysUtil.is_package_installed("nonexistent_package"))
        with patch("importlib.util.find_spec", return_value=True):
            self.assertTrue(SysUtil.is_package_installed("existent_package"))

    @patch(
        "importlib.metadata.distributions",
        return_value=[type("", (), {"metadata": {"Name": "fake-package"}})()],
    )
    def test_list_installed_packages(self, mock_distributions):
        installed_packages = SysUtil.list_installed_packages()
        self.assertIn("fake-package", installed_packages)

    @patch("subprocess.check_call")
    def test_uninstall_package(self, mock_subprocess):
        SysUtil.uninstall_package("fake-package")
        mock_subprocess.assert_called_with(
            [sys.executable, "-m", "pip", "uninstall", "fake-package", "-y"]
        )

    @patch("subprocess.check_call")
    def test_update_package(self, mock_subprocess):
        SysUtil.update_package("fake-package")
        mock_subprocess.assert_called_with(
            [sys.executable, "-m", "pip", "install", "--upgrade", "fake-package"]
        )

    def test_split_path_file(self):
        """Test splitting a file path."""
        parent, name = SysUtil.split_path("/tmp/example.txt")
        self.assertEqual(parent, Path("/tmp"))
        self.assertEqual(name, "example.txt")

    def test_split_path_directory(self):
        """Test splitting a directory path."""
        parent, name = SysUtil.split_path("/tmp/example/")
        self.assertEqual(parent, Path("/tmp"))
        self.assertEqual(name, "example")

    def test_split_path_root(self):
        """Test splitting the root path."""
        parent, name = SysUtil.split_path("/")
        self.assertEqual(parent, Path("/"))
        self.assertEqual(name, "")

    def setUp(self):
        # Create a temporary directory
        self.test_dir = tempfile.mkdtemp()

        # Create some test files
        Path(self.test_dir, "test_file.txt").touch()
        Path(self.test_dir, "another_test_file.md").touch()

    def tearDown(self):
        # Remove the temporary directory after the test
        shutil.rmtree(self.test_dir)

    def test_list_files_no_extension(self):
        files = SysUtil.list_files(self.test_dir)
        self.assertEqual(len(files), 2)  # Expecting 2 files

    def test_list_files_with_extension(self):
        files = SysUtil.list_files(self.test_dir, extension="txt")
        self.assertEqual(len(files), 1)  # Expecting 1 .txt file
        self.assertTrue(files[0].name.endswith(".txt"))

    def test_change_existing_key(self):
        test_dict = {"old_key": "value"}
        SysUtil.change_dict_key(test_dict, "old_key", "new_key")
        self.assertIn("new_key", test_dict)
        self.assertNotIn("old_key", test_dict)

    def test_change_non_existing_key(self):
        test_dict = {"key": "value"}
        SysUtil.change_dict_key(test_dict, "non_existing_key", "new_key")
        self.assertNotIn("new_key", test_dict)
        self.assertIn("key", test_dict)

    def test_change_key_to_existing_key(self):
        test_dict = {"old_key": "value1", "new_key": "value2"}
        SysUtil.change_dict_key(test_dict, "old_key", "new_key")
        self.assertEqual(test_dict["new_key"], "value1")

    def test_default_separator(self):
        timestamp = SysUtil.get_timestamp()
        self.assertNotIn(":", timestamp)
        self.assertNotIn(".", timestamp)

    def test_custom_separator(self):
        timestamp = SysUtil.get_timestamp(sep="-")
        self.assertNotIn(":", timestamp)
        self.assertNotIn(".", timestamp)
        self.assertIn("-", timestamp)

    def test_valid_schema(self):
        test_dict = {"key1": 1, "key2": "value"}
        schema = {"key1": int, "key2": str}
        self.assertTrue(SysUtil.is_schema(test_dict, schema))

    def test_invalid_schema(self):
        test_dict = {"key1": "value", "key2": "value"}
        schema = {"key1": int, "key2": str}
        self.assertFalse(SysUtil.is_schema(test_dict, schema))

    def test_partial_schema(self):
        test_dict = {"key1": 1}
        schema = {"key1": int, "key2": str}
        self.assertFalse(SysUtil.is_schema(test_dict, schema))

    def test_single_copy(self):
        original = {"key": "value"}
        copy = SysUtil.create_copy(original)
        self.assertEqual(original, copy)
        self.assertNotEqual(id(original), id(copy))

    def test_multiple_copies(self):
        original = {"key": "value"}
        copies = SysUtil.create_copy(original, num=2)
        self.assertEqual(len(copies), 2)
        self.assertNotEqual(id(copies[0]), id(copies[1]))

    def test_invalid_num(self):
        with self.assertRaises(ValueError):
            SysUtil.create_copy({}, num=0)

    def test_id_length(self):
        id_ = SysUtil.create_id()
        self.assertEqual(len(id_), 32)

    def test_custom_length(self):
        id_ = SysUtil.create_id(n=16)
        self.assertEqual(len(id_), 16)

    def test_bins_with_small_strings(self):
        input_ = ["a", "b", "c", "d"]
        bins = SysUtil.get_bins(input_, upper=2)
        self.assertEqual(len(bins), 4)

    def test_bins_with_large_strings(self):
        input_ = ["a" * 1000, "b" * 1000, "c" * 1000]
        bins = SysUtil.get_bins(input_)
        self.assertEqual(len(bins), 3)

    def test_bins_with_empty_input(self):
        bins = SysUtil.get_bins([])
        self.assertEqual(len(bins), 0)


if __name__ == "__main__":
    unittest.main()



"""
Module for base component model definition using Pydantic.
"""

from abc import ABC
from typing import Any, TypeVar

from pydantic import Field, field_serializer, AliasChoices
from lionagi.libs import SysUtil, convert

from .base_mixin import BaseComponentMixin

T = TypeVar("T", bound="BaseComponent")


class BaseComponent(BaseComponentMixin, ABC):
    """
    A base component model that provides common attributes and utility methods for metadata management.
    It includes functionality to interact with metadata in various ways, such as retrieving, modifying,
    and validating metadata keys and values.

    Attributes:
            id_ (str): Unique identifier, defaulted using SysUtil.create_id.
            timestamp (str | None): Timestamp of creation or modification.
            metadata (dict[str, Any]): Metadata associated with the component.
    """

    id_: str = Field(default_factory=SysUtil.create_id, alias="node_id")
    timestamp: str | None = Field(default_factory=SysUtil.get_timestamp)
    metadata: dict[str, Any] = Field(default_factory=dict, alias="meta")

    class Config:
        """Model configuration settings."""

        extra = "allow"
        arbitrary_types_allowed = True
        populate_by_name = True
        validate_assignment = True
        validate_return = True
        str_strip_whitespace = True

    @classmethod
    def class_name(cls) -> str:
        """
        Retrieves the name of the class.
        """
        return cls.__name__

    @property
    def property_schema(self):
        return self.model_json_schema()["properties"]

    @property
    def property_keys(self):
        return list(self.model_json_schema()["properties"].keys())

    def copy(self, *args, **kwargs) -> T:
        """
        Creates a deep copy of the instance, with an option to update specific fields.

        Args:
                *args: Variable length argument list for additional options.
                **kwargs: Arbitrary keyword arguments specifying updates to the instance.

        Returns:
                BaseComponent: A new instance of BaseComponent as a deep copy of the original, with updates applied.
        """
        return self.model_copy(*args, **kwargs)

    def __repr__(self):
        return f"{self.__class__.__name__}({self.to_dict()})"


class BaseNode(BaseComponent):
    """
    A base class for nodes, representing a fundamental unit in a graph or tree structure,
    extending BaseComponent with content handling capabilities.

    Attributes:
            content: The content of the node, which can be a string, a dictionary with any structure,
                    None, or any other type. It is flexible to accommodate various types of content.
                    This attribute also supports aliasing through validation_alias for compatibility with
                    different naming conventions like "text", "page_content", or "chunk_content".
    """

    content: str | dict[str, Any] | None | Any = Field(
        default=None,
        validation_alias=AliasChoices("text", "page_content", "chunk_content"),
    )

    @property
    def content_str(self):
        """
        Attempts to serialize the node's content to a string.

        Returns:
                str: The serialized content string. If serialization fails, returns "null" and
                        logs an error message indicating the content is not serializable.
        """
        try:
            return convert.to_str(self.content)
        except ValueError:
            print(
                f"Content is not serializable for Node: {self._id}, defaulting to 'null'"
            )
            return "null"

    def __str__(self):
        """
        Provides a string representation of the BaseNode instance, including a content preview,
        metadata preview, and optionally the timestamp if present.

        Returns:
                str: A string representation of the instance.
        """
        timestamp = f" ({self.timestamp})" if self.timestamp else ""
        if self.content:
            content_preview = (
                f"{self.content[:50]}..." if len(self.content) > 50 else self.content
            )
        else:
            content_preview = ""
        meta_preview = (
            f"{str(self.metadata)[:50]}..."
            if len(str(self.metadata)) > 50
            else str(self.metadata)
        )
        return (
            f"{self.class_name()}({self.id_}, {content_preview}, {meta_preview},"
            f"{timestamp})"
        )


class BaseRelatableNode(BaseNode):
    """
    Extends BaseNode with functionality to manage relationships with other nodes.

    Attributes:
            related_nodes: A list of identifiers (str) for nodes that are related to this node.
            label: An optional label for the node, providing additional context or classification.
    """

    related_nodes: list[str] = Field(default_factory=list)
    label: str | None = None

    def add_related_node(self, node_id: str) -> bool:
        """
        Adds a node to the list of related nodes if it's not already present.

        Args:
                node_id: The identifier of the node to add.

        Returns:
                bool: True if the node was added, False if it was already in the list.
        """
        if node_id not in self.related_nodes:
            self.related_nodes.append(node_id)
            return True
        return False

    def remove_related_node(self, node_id: str) -> bool:
        """
        Removes a node from the list of related nodes if it's present.

        Args:
                node_id: The identifier of the node to remove.

        Returns:
                bool: True if the node was removed, False if it was not found in the list.
        """

        if node_id in self.related_nodes:
            self.related_nodes.remove(node_id)
            return True
        return False


class Tool(BaseRelatableNode):
    """
    Represents a tool, extending BaseRelatableNode with specific functionalities and configurations.

    Attributes:
            func: The main function or capability of the tool.
            schema_: An optional schema defining the structure and constraints of data the tool works with.
            manual: Optional documentation or manual for using the tool.
            parser: An optional parser associated with the tool for data processing or interpretation.
    """

    func: Any
    schema_: dict | None = None
    manual: Any | None = None
    parser: Any | None = None

    @field_serializer("func")
    def serialize_func(self, func):
        return func.__name__


TOOL_TYPE = bool | Tool | str | list[Tool | str | dict] | dict



import setuptools
import os

here = os.path.abspath(os.path.dirname(__file__))

# Get the code version
version = {}
with open(os.path.join(here, "lionagi/version.py")) as fp:
    exec(fp.read(), version)
__version__ = version["__version__"]

install_requires = [
    "aiohttp",
    "python-dotenv",
    "tiktoken",
    "pydantic",
    "aiocache",
    "pandas",
]

setuptools.setup(
    name="lionagi",
    version=__version__,
    author="HaiyangLi",
    author_email="ocean@lionagi.ai",
    description="Towards automated general intelligence.",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(include=["lionagi*"]),
    install_requires=install_requires,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.10",
)



# Default configs for the OpenAI API

# ChatCompletion
oai_chat_llmconfig = {
    "model": "gpt-4-turbo-preview",
    "frequency_penalty": 0,
    "max_tokens": None,
    "n": 1,
    "presence_penalty": 0,
    "response_format": {"type": "text"},
    "seed": None,
    "stop": None,
    "stream": False,
    "temperature": 0.7,
    "top_p": 1,
    "tools": None,
    "tool_choice": "none",
    "user": None,
}

oai_chat_schema = {
    "required": [
        "model",
        "frequency_penalty",
        "n",
        "presence_penalty",
        "response_format",
        "temperature",
        "top_p",
    ],
    "optional": [
        "seed",
        "stop",
        "stream",
        "tools",
        "tool_choice",
        "user",
        "max_tokens",
    ],
    "input_": "messages",
    "config": oai_chat_llmconfig,
}

# Finetune
oai_finetune_llmconfig = {
    "model": "gpt-3.5-turbo",
    "hyperparameters": {
        "batch_size": "auto",
        "learning_rate_multiplier": "auto",
        "n_epochs": "auto",
    },
    "suffix": None,
    "training_file": None,
}

oai_finetune_schema = {
    "required": ["model", "training_file"],
    "optional": ["hyperparameters", "suffix", "validate_file"],
    "input_": ["training_file"],
    "config": oai_finetune_llmconfig,
}

# Audio ---- create  speech

oai_audio_speech_llmconfig = {
    "model": "tts-1",
    "voice": "alloy",
    "response_format": "mp3",
    "speed": 1,
}
oai_audio_speech_schema = {
    "required": ["model", "voice"],
    "optional": ["response_format", "speed"],
    "input_": "input_",
    "config": oai_audio_speech_llmconfig,
}

# Audio ----------- create transcription
oai_audio_transcriptions_llmconfig = {
    "model": "whisper-1",
    "language": None,
    "prompt": None,
    "response_format": "json",
    "temperature": 0,
}
oai_audio_transcriptions_schema = {
    "required": ["model", "voice"],
    "optional": [
        "response_format",
        "language",
        "prompt",
        "response_format",
        "temperature",
    ],
    "input_": "file",
    "config": oai_audio_transcriptions_llmconfig,
}

# Audio ------------    translations
oai_audio_translations_llmconfig = {
    "model": "whisper-1",
    "prompt": None,
    "response_format": "json",
    "temperature": 0,
}

oai_audio_translations_schema = {
    "required": ["model"],
    "optional": ["response_format", "speed", "prompt", "temperature"],
    "input_": "file",
    "config": oai_audio_translations_llmconfig,
}

# images


oai_schema = {
    "chat/completions": oai_chat_schema,
    "finetune": oai_finetune_schema,
    "audio_speech": oai_audio_speech_schema,
    "audio_transcriptions": oai_audio_transcriptions_schema,
    "audio_translations": oai_audio_translations_schema,
}



from lionagi.libs.ln_api import BaseService
from lionagi.integrations.config.ollama_configs import model

allowed_kwargs = [
    "model",
    "frequency_penalty",
    "max_tokens",
    "n",
    "presence_penalty",
    "response_format",
    "seed",
    "stop",
    "stream",
    "temperature",
    "top_p",
    "tools",
    "tool_choice",
    "user",
]


class OllamaService(BaseService):

    def __init__(self, model: str = model, **kwargs):
        super().__init__()

        from lionagi.libs.sys_util import SysUtil

        SysUtil.check_import("ollama")

        import ollama

        self.ollama = ollama
        self.model = model
        self.client = self.ollama.AsyncClient(**kwargs)

    async def serve_chat(self, messages, **kwargs):
        config = {}
        for k, v in kwargs.items():
            if k in allowed_kwargs:
                config[k] = v

        self.ollama.pull(self.model)
        payload = {"messages": messages}

        try:
            completion = await self.client.chat(
                model=self.model, messages=messages, **config
            )
            completion["choices"] = [{"message": completion.pop("message")}]
            return payload, completion
        except Exception as e:
            self.status_tracker.num_tasks_failed += 1
            raise e



"""
This module contains the Branch class, which represents a branch in a conversation tree.
"""

from collections import deque
from typing import Any, Union, TypeVar, Callable

from lionagi.libs.sys_util import PATH_TYPE
from lionagi.libs import StatusTracker, BaseService, convert, dataframe

from ..schema import TOOL_TYPE, Tool, DataLogger
from ..tool import ToolManager, func_to_tool

from ..messages import System
from ..mail import BaseMail

from .util import MessageUtil
from .base_branch import BaseBranch
from .branch_flow_mixin import BranchFlowMixin

from dotenv import load_dotenv

load_dotenv()

T = TypeVar("T", bound=Tool)


class Branch(BaseBranch, BranchFlowMixin):
    """
    Represents a branch in a conversation tree.

    Attributes:
        sender (str): The sender of the branch (default: "system").
        tool_manager (ToolManager): The tool manager for the branch.
        service (BaseService): The service associated with the branch.
        llmconfig (dict): The configuration for the language model.
        status_tracker (StatusTracker): The status tracker for the branch.
        pending_ins (dict): The pending incoming mails for the branch.
        pending_outs (deque): The pending outgoing mails for the branch.

    Methods:
        __init__(self, name=None, system=None, messages=None, service=None, sender=None,
                 llmconfig=None, tools=None, datalogger=None, persist_path=None,
                 tool_manager=None, **kwargs) -> None:
            Initializes the Branch instance.

        from_csv(cls, filepath, name=None, service=None, llmconfig=None, tools=None,
                 datalogger=None, persist_path=None, tool_manager=None, read_kwargs=None,
                 **kwargs) -> Branch:
            Creates a Branch instance from a CSV file.

        from_json_string(cls, filepath, name=None, service=None, llmconfig=None, tools=None,
                         datalogger=None, persist_path=None, tool_manager=None, read_kwargs=None,
                         **kwargs) -> Branch:
            Creates a Branch instance from a JSON string file.

        messages_describe(self) -> dict[str, Any]:
            Describes the messages in the branch.

        has_tools(self) -> bool:
            Checks if the branch has any registered tools.

        merge_branch(self, branch: Branch, update=True) -> None:
            Merges another branch into the current branch.

        register_tools(self, tools: Union[Tool, list[Tool]]) -> None:
            Registers tools in the branch's tool manager.

        delete_tools(self, tools: Union[T, list[T], str, list[str]], verbose=True) -> bool:
            Deletes tools from the branch's tool manager.

        send(self, recipient: str, category: str, package: Any) -> None:
            Sends a mail to a recipient.

        receive(self, sender: str, messages=True, tools=True, service=True, llmconfig=True) -> None:
            Receives mails from a sender and updates the branch accordingly.

        receive_all(self) -> None:
            Receives all pending mails and updates the branch accordingly.

        _add_service(service, llmconfig) -> tuple[BaseService, dict]:
            Adds a service and its configuration to the branch.

        _is_invoked(self) -> bool:
            Checks if the conversation has been invoked with an action response.
    """

    def __init__(
        self,
        name: str | None = None,
        system: dict | list | System | None = None,
        messages: dataframe.ln_DataFrame | None = None,
        service: BaseService | None = None,
        sender: str | None = None,
        llmconfig: dict[str, str | int | dict] | None = None,
        tools: list[Callable | Tool] | None = None,
        datalogger: None | DataLogger = None,
        persist_path: PATH_TYPE | None = None,  # instruction_sets=None,
        tool_manager: ToolManager | None = None,
        **kwargs,
    ):
        """
        Initializes the Branch instance.

        Args:
            name (str): The name of the branch (optional).
            system (dict | list | System): The system message for the branch (optional).
            messages (dataframe.ln_DataFrame): The messages in the branch (optional).
            service (BaseService): The service associated with the branch (optional).
            sender (str): The sender of the branch (optional, default: "system").
            llmconfig (dict[str, str | int | dict]): The configuration for the language model (optional).
            tools (list[Callable | Tool]): The tools to register in the branch (optional).
            datalogger (DataLogger): The data logger for the branch (optional).
            persist_path (PATH_TYPE): The path to persist the branch data (optional).
            tool_manager (ToolManager): The tool manager for the branch (optional).
            **kwargs: Additional keyword arguments.

        Raises:
            TypeError: If there is an error in registering the tools.
        """

        super().__init__(
            messages=messages,
            datalogger=datalogger,
            persist_path=persist_path,
            name=name,
            **kwargs,
        )

        self.sender = sender or "system"
        self.tool_manager = tool_manager or ToolManager()

        if tools:
            try:
                tools_ = []
                _tools = convert.to_list(tools)
                for i in _tools:
                    if isinstance(i, Tool):
                        tools_.append(i)
                    else:
                        tools_.append(func_to_tool(i))

                self.register_tools(tools_)
            except Exception as e:
                raise TypeError(f"Error in registering tools: {e}") from e

        self.service, self.llmconfig = self._add_service(service, llmconfig)
        self.status_tracker = StatusTracker()

        # add instruction sets
        # self.instruction_sets = instruction_sets

        self.pending_ins = {}
        self.pending_outs = deque()

        if system is not None:
            self.add_message(system=system)

    @classmethod
    def from_csv(
        cls,
        filepath,
        name: str | None = None,
        service: BaseService | None = None,
        llmconfig: dict[str, str | int | dict] | None = None,
        tools: TOOL_TYPE | None = None,
        datalogger: None | DataLogger = None,
        persist_path: PATH_TYPE | None = None,  # instruction_sets=None,
        tool_manager: ToolManager | None = None,
        read_kwargs=None,
        **kwargs,
    ) -> "Branch":
        """
        Creates a Branch instance from a CSV file.

        Args:
            filepath: The path to the CSV file.
            name (str): The name of the branch (optional).
            service (BaseService): The service associated with the branch (optional).
            llmconfig (dict[str, str | int | dict]): The configuration for the language model (optional).
            tools (TOOL_TYPE): The tools to register in the branch (optional).
            datalogger (DataLogger): The data logger for the branch (optional).
            persist_path (PATH_TYPE): The path to persist the branch data (optional).
            tool_manager (ToolManager): The tool manager for the branch (optional).
            read_kwargs: Additional keyword arguments for reading the CSV file (optional).
            **kwargs: Additional keyword arguments.

        Returns:
            Branch: The created Branch instance.
        """
        return cls._from_csv(
            filepath=filepath,
            read_kwargs=read_kwargs,
            name=name,
            service=service,
            llmconfig=llmconfig,
            tools=tools,
            datalogger=datalogger,
            persist_path=persist_path,
            # instruction_sets=instruction_sets,
            tool_manager=tool_manager,
            **kwargs,
        )

    @classmethod
    def from_json_string(
        cls,
        filepath,
        name: str | None = None,
        service: BaseService | None = None,
        llmconfig: dict[str, str | int | dict] | None = None,
        tools: TOOL_TYPE | None = None,
        datalogger: None | DataLogger = None,
        persist_path: PATH_TYPE | None = None,  # instruction_sets=None,
        tool_manager: ToolManager | None = None,
        read_kwargs=None,
        **kwargs,
    ) -> "Branch":
        """
        Creates a Branch instance from a JSON string file.

        Args:
            filepath: The path to the JSON string file.
            name (str): The name of the branch (optional).
            service (BaseService): The service associated with the branch (optional).
            llmconfig (dict[str, str | int | dict]): The configuration for the language model (optional).
            tools (TOOL_TYPE): The tools to register in the branch (optional).
            datalogger (DataLogger): The data logger for the branch (optional).
            persist_path (PATH_TYPE): The path to persist the branch data (optional).
            tool_manager (ToolManager): The tool manager for the branch (optional).
            read_kwargs: Additional keyword arguments for reading the JSON string file (optional).
            **kwargs: Additional keyword arguments.

        Returns:
            Branch: The created Branch instance.
        """
        return cls._from_json(
            filepath=filepath,
            read_kwargs=read_kwargs,
            name=name,
            service=service,
            llmconfig=llmconfig,
            tools=tools,
            datalogger=datalogger,
            persist_path=persist_path,
            # instruction_sets=instruction_sets,
            tool_manager=tool_manager,
            **kwargs,
        )

    def messages_describe(self) -> dict[str, Any]:
        """
        Describes the messages in the branch.

        Returns:
            dict[str, Any]: A dictionary describing the messages in the branch.
        """
        return dict(
            total_messages=len(self.messages),
            summary_by_role=self._info(),
            summary_by_sender=self._info(use_sender=True),
            # instruction_sets=self.instruction_sets,
            registered_tools=self.tool_manager.registry,
            messages=[msg.to_dict() for _, msg in self.messages.iterrows()],
        )

    @property
    def has_tools(self) -> bool:
        """
        Checks if the branch has any registered tools.

        Returns:
            bool: True if the branch has registered tools, False otherwise.
        """
        return self.tool_manager.registry != {}

    # todo: also update other attributes
    def merge_branch(self, branch: "Branch", update: bool = True) -> None:
        """
        Merges another branch into the current branch.

        Args:
            branch (Branch): The branch to merge.
            update (bool): Whether to update the existing attributes or add new ones (default: True).
        """
        message_copy = branch.messages.copy()
        self.messages = self.messages.merge(message_copy, how="outer")
        self.datalogger.extend(branch.datalogger.log)

        if update:
            # self.instruction_sets.update(branch.instruction_sets)
            self.tool_manager.registry.update(branch.tool_manager.registry)
        else:
            for key, value in branch.instruction_sets.items():
                if key not in self.instruction_sets:
                    self.instruction_sets[key] = value

            for key, value in branch.tool_manager.registry.items():
                if key not in self.tool_manager.registry:
                    self.tool_manager.registry[key] = value

    # ----- tool manager methods ----- #
    def register_tools(
        self, tools: Union[Tool, list[Tool | Callable], Callable]
    ) -> None:
        """
        Registers tools in the branch's tool manager.

        Args:
            tools (Union[Tool, list[Tool]]): The tool(s) to register.
        """
        if not isinstance(tools, list):
            tools = [tools]
        self.tool_manager.register_tools(tools=tools)

    def delete_tools(
        self,
        tools: Union[T, list[T], str, list[str]],
        verbose: bool = True,
    ) -> bool:
        """
        Deletes tools from the branch's tool manager.

        Args:
            tools (Union[T, list[T], str, list[str]]): The tool(s) to delete.
            verbose (bool): Whether to print success/failure messages (default: True).

        Returns:
            bool: True if the tools were successfully deleted, False otherwise.
        """
        if not isinstance(tools, list):
            tools = [tools]
        if convert.is_same_dtype(tools, str):
            for act_ in tools:
                if act_ in self.tool_manager.registry:
                    self.tool_manager.registry.pop(act_)
            if verbose:
                print("tools successfully deleted")
            return True
        elif convert.is_same_dtype(tools, Tool):
            for act_ in tools:
                if act_.schema_["function"]["name"] in self.tool_manager.registry:
                    self.tool_manager.registry.pop(act_.schema_["function"]["name"])
            if verbose:
                print("tools successfully deleted")
            return True
        if verbose:
            print("tools deletion failed")
        return False

    def send(self, recipient: str, category: str, package: Any) -> None:
        """
        Sends a mail to a recipient.

        Args:
            recipient (str): The recipient of the mail.
            category (str): The category of the mail.
            package (Any): The package to send in the mail.
        """
        mail_ = BaseMail(
            sender=self.sender, recipient=recipient, category=category, package=package
        )
        self.pending_outs.append(mail_)

    def receive(
        self,
        sender: str,
        messages: bool = True,
        tools: bool = True,
        service: bool = True,
        llmconfig: bool = True,
    ) -> None:
        """
        Receives mails from a sender and updates the branch accordingly.

        Args:
            sender (str): The sender of the mails.
            messages (bool): Whether to receive message updates (default: True).
            tools (bool): Whether to receive tool updates (default: True).
            service (bool): Whether to receive service updates (default: True).
            llmconfig (bool): Whether to receive language model configuration updates (default: True).

        Raises:
            ValueError: If there are no packages from the specified sender.
                        If the messages format is invalid.
                        If the tools format is invalid.
                        If the provider format is invalid.
                        If the llmconfig format is invalid.
        """
        skipped_requests = deque()
        if sender not in self.pending_ins:
            raise ValueError(f"No package from {sender}")
        while self.pending_ins[sender]:
            mail_ = self.pending_ins[sender].popleft()

            if mail_.category == "messages" and messages:
                if not isinstance(mail_.package, dataframe.ln_DataFrame):
                    raise ValueError("Invalid messages format")
                MessageUtil.validate_messages(mail_.package)
                self.messages = self.messages.merge(mail_.package, how="outer")

            elif mail_.category == "tools" and tools:
                if not isinstance(mail_.package, Tool):
                    raise ValueError("Invalid tools format")
                self.tool_manager.register_tools([mail_.package])

            elif mail_.category == "provider" and service:
                from lionagi.libs.ln_api import BaseService

                if not isinstance(mail_.package, BaseService):
                    raise ValueError("Invalid provider format")
                self.service = mail_.package

            elif mail_.category == "llmconfig" and llmconfig:
                if not isinstance(mail_.package, dict):
                    raise ValueError("Invalid llmconfig format")
                self.llmconfig.update(mail_.package)

            else:
                skipped_requests.append(mail_)

        self.pending_ins[sender] = skipped_requests
        if self.pending_ins[sender] == deque():
            self.pending_ins.pop(sender)

    def receive_all(self) -> None:
        """
        Receives all pending mails and updates the branch accordingly.
        """
        for key in list(self.pending_ins.keys()):
            self.receive(key)

    @staticmethod
    def _add_service(service, llmconfig):
        from lionagi.integrations.provider.oai import OpenAIService

        if service is None:
            try:
                from lionagi.integrations.provider import Services

                service = Services.OpenAI()

            except:
                raise ValueError("No available service")
        if llmconfig is None:
            if isinstance(service, OpenAIService):
                from lionagi.integrations.config import oai_schema

                llmconfig = oai_schema["chat/completions"]["config"]
            else:
                llmconfig = {}
        return service, llmconfig

    def _is_invoked(self) -> bool:
        """
        Check if the conversation has been invoked with an action response.

        Returns:
                bool: True if the conversation has been invoked, False otherwise.

        """
        content = self.messages.iloc[-1]["content"]
        try:
            if convert.to_dict(content)["action_response"].keys() >= {
                "function",
                "arguments",
                "output",
            }:
                return True
        except Exception:
            return False



from typing import Any, TypeVar
from lionagi.libs.sys_util import SysUtil


def to_llama_index_node(lion_node, node_type: Any = None, **kwargs: Any) -> Any:
    """
    Converts a Lion node to a Llama Index node of a specified type.

    This function takes a node from the Lion framework, converts it to a dictionary, modifies keys to match
    the expected Llama Index node schema, and then creates a Llama Index node object of the specified type.

    Args:
            lion_node: The Lion node to convert. Must have a `to_dict` method.
            node_type (Any, optional): The type of Llama Index node to create. Can be a string name of a node class
                    within the Llama Index schema or a class that inherits from `BaseNode`. Defaults to 'TextNode'.
            **kwargs: Additional keyword arguments to be included in the Llama Index node's initialization.

    Returns:
            Any: A new instance of the specified Llama Index node type populated with data from the Lion node.

    Raises:
            TypeError: If `node_type` is neither a string nor a subclass of `BaseNode`.
            AttributeError: If an error occurs due to an invalid node type or during the creation of the node object.
    """

    SysUtil.check_import("llama_index", pip_name="llama-index")
    from llama_index.core.schema import BaseNode
    import llama_index.core.schema

    node_type = node_type or "TextNode"

    _dict = lion_node.to_dict()
    SysUtil.change_dict_key(_dict, old_key="content", new_key="text")
    SysUtil.change_dict_key(_dict, old_key="node_id", new_key="id_")
    _dict["text"] = str(_dict["text"])
    _dict = {**_dict, **kwargs}

    if not isinstance(node_type, str) and not issubclass(node_type, BaseNode):
        raise TypeError(f"node_type must be a string or BaseNode")

    else:
        try:
            if isinstance(node_type, str) and hasattr(
                llama_index.core.schema, node_type
            ):
                return getattr(llama_index.core.schema, node_type).from_dict(_dict)
            elif issubclass(node_type, BaseNode):
                return node_type.from_dict(_dict)
            else:
                raise AttributeError(f"Invalid llama-index node type: {node_type}")
        except Exception as e:
            raise AttributeError(f"Error: {e}")



from typing import Any
from .base_node import BaseNode

from lionagi.integrations.bridge import LlamaIndexBridge, LangchainBridge


class DataNode(BaseNode):
    """
    Represents a data node with extended functionality for integration with llama index and langchain_ formats.

    This class extends `BaseNode` to include methods for converting between DataNode instances and specific formats
    used by llama index and langchain_, facilitating interoperability with these systems.

    Methods provided allow for serialization to and deserialization from these formats, supporting a variety of use cases.
    """

    def to_llama_index(self, node_type=None, **kwargs) -> Any:
        """
        Converts the node to a format compatible with llama index.

        This method serializes the DataNode into a format recognized by the llama index system, allowing for
        integration and usage within that ecosystem.

        Args:
                node_type:
                **kwargs: Additional keyword arguments for customization.

        Returns:
                Any: The llama index format representation of the node.

        Examples:
                >>> node = DataNode(content="Example content")
                >>> llama_index = node.to_llama_index()
        """
        return LlamaIndexBridge.to_llama_index_node(self, node_type=node_type, **kwargs)

    def to_langchain(self, **kwargs) -> Any:
        """
        Converts the node to a langchain_ document format.

        This method serializes the DataNode into a document format used by langchain_, enabling the node's
        use within langchain_ applications and workflows.

        Args:
                **kwargs: Additional keyword arguments for customization.

        Returns:
                Any: The langchain_ document representation of the node.

        Examples:
                >>> node = DataNode(content="Example content")
                >>> langchain_doc = node.to_langchain()
        """
        return LangchainBridge.to_langchain_document(self, **kwargs)

    @classmethod
    def from_llama_index(cls, llama_node: Any, **kwargs) -> "DataNode":
        """
        Creates a DataNode instance from a llama index node.

        Args:
                llama_node: The llama index node object.
                **kwargs: Variable length argument list.

        Returns:
                An instance of DataNode.

        Examples:
                llama_node = SomeLlamaIndexNode()
                data_node = DataNode.from_llama_index(llama_node)
        """
        llama_dict = llama_node.to_dict(**kwargs)
        return cls.from_obj(llama_dict)

    @classmethod
    def from_langchain(cls, lc_doc: Any) -> "DataNode":
        """
        Creates a DataNode instance from a langchain_ document.

        Args:
                lc_doc: The langchain_ document object.

        Returns:
                An instance of DataNode.

        Examples:
                lc_doc = SomeLangChainDocument()
                data_node = DataNode.from_langchain(lc_doc)
        """
        info_json = lc_doc.to_json()
        info_node = {"lc_id": info_json["id"]}
        info_node = {**info_node, **info_json["kwargs"]}
        return cls(**info_node)



"""
This module contains the ScoreTemplate class and related functions for scoring a given context using a language model.

The ScoreTemplate class is a subclass of ScoredTemplate and provides functionality for scoring a given context
based on specified instructions, score range, and other parameters. It includes fields for the input sentence,
score range, inclusive flag, number of digits, confidence score, and reason for the score.

The module also includes functions for scoring a single instance or multiple instances of the context using the
ScoreTemplate class and a language model.
"""

from pydantic import Field
import numpy as np
from lionagi.libs import func_call, convert
from ..prompt.prompt_template import ScoredTemplate
from ..branch import Branch


class ScoreTemplate(ScoredTemplate):
    """
    A class for scoring a given context using a language model.

    Attributes:
        template_name (str): The name of the score template (default: "default_score").
        sentence (str | list | dict): The given context to score.
        answer (float): The numeric score for the context.
        signature (str): The signature indicating the input and output fields (default: "sentence -> answer").

    Methods:
        __init__(self, sentence=None, instruction=None, score_range=(1, 10), inclusive=True, num_digit=0,
                 confidence_score=False, reason=False, **kwargs):
            Initializes a new instance of the ScoreTemplate class.
    """

    template_name: str = "default_score"
    sentence: str | list | dict = Field(
        default_factory=str, description="the given context to score"
    )
    answer: float = Field(default_factory=float, description=f"a numeric score")
    signature: str = "sentence -> answer"

    def __init__(
        self,
        sentence=None,
        instruction=None,
        score_range=(1, 10),
        inclusive=True,
        num_digit=0,
        confidence_score=False,
        reason=False,
        **kwargs,
    ):
        super().__init__(**kwargs)

        self.sentence = sentence

        return_precision = ""
        if num_digit == 0:
            return_precision = "integer"
        else:
            return_precision = f"num:{convert.to_str(num_digit)}f"

        self.task = f"""
score context according to the following constraints
1. objective, {convert.to_str(instruction)}
2. score range, {convert.to_str(score_range)}
3. include_endpoints, {"yes" if inclusive else "no"}
4. format the score in {return_precision}
"""

        if reason:
            self.output_fields.append("reason")

        if confidence_score:
            self.output_fields.append("confidence_score")

        self.out_validation_kwargs["answer"] = {
            "upper_bound": score_range[1],
            "lower_bound": score_range[0],
            "num_type": int if num_digit == 0 else float,
            "precision": num_digit if num_digit != 0 else None,
        }


async def _score(
    sentence,
    instruction=None,
    score_range=(1, 10),
    inclusive=True,
    num_digit=0,
    confidence_score=False,
    reason=False,
    retries=2,
    delay=0.5,
    backoff_factor=2,
    default_value=None,
    timeout=None,
    branch_name=None,
    system=None,
    messages=None,
    service=None,
    sender=None,
    llmconfig=None,
    tools=None,
    datalogger=None,
    persist_path=None,
    tool_manager=None,
    **kwargs,
):
    """
    Scores a given context using a language model.

    Args:
        sentence (str | list | dict): The given context to score.
        instruction (Optional[str]): The instruction for scoring the context.
        score_range (tuple): The range of valid scores (default: (1, 10)).
        inclusive (bool): Whether to include the endpoints of the score range (default: True).
        num_digit (int): The number of digits after the decimal point for the score (default: 0).
        confidence_score (bool): Whether to include the confidence score in the output (default: False).
        reason (bool): Whether to include the reason for the score in the output (default: False).
        retries (int): The number of retries for the API call (default: 2).
        delay (float): The initial delay between retries in seconds (default: 0.5).
        backoff_factor (float): The backoff factor for exponential delay between retries (default: 2).
        default_value (Optional[Any]): The default value to return if the API call fails (default: None).
        timeout (Optional[float]): The timeout for the API call in seconds (default: None).
        branch_name (Optional[str]): The name of the branch to use for scoring.
        system (Optional[Any]): The system configuration for the branch.
        messages (Optional[Any]): The messages to initialize the branch with.
        service (Optional[Any]): The service to use for scoring.
        sender (Optional[str]): The sender of the scoring request.
        llmconfig (Optional[Any]): The configuration for the language model.
        tools (Optional[Any]): The tools to use for scoring.
        datalogger (Optional[Any]): The data logger for the branch.
        persist_path (Optional[str]): The path to persist the branch data.
        tool_manager (Optional[Any]): The tool manager for the branch.
        **kwargs: Additional keyword arguments for the API call.

    Returns:
        ScoreTemplate: The score template with the scored context.
    """

    if "temperature" not in kwargs:
        kwargs["temperature"] = 0.1

    instruction = instruction or ""

    branch = Branch(
        name=branch_name,
        system=system,
        messages=messages,
        service=service,
        sender=sender,
        llmconfig=llmconfig,
        tools=tools,
        datalogger=datalogger,
        persist_path=persist_path,
        tool_manager=tool_manager,
    )

    _template = ScoreTemplate(
        sentence=sentence,
        instruction=instruction,
        score_range=score_range,
        inclusive=inclusive,
        num_digit=num_digit,
        confidence_score=confidence_score,
        reason=reason,
    )

    await func_call.rcall(
        branch.chat,
        prompt_template=_template,
        retries=retries,
        delay=delay,
        backoff_factor=backoff_factor,
        default=default_value,
        timeout=timeout,
        **kwargs,
    )

    return _template


async def score(
    sentence,
    num_instances=1,
    instruction=None,
    score_range=(1, 10),
    inclusive=True,
    num_digit=0,
    confidence_score=False,
    reason=False,
    retries=2,
    delay=0.5,
    backoff_factor=2,
    default_value=None,
    timeout=None,
    branch_name=None,
    system=None,
    messages=None,
    service=None,
    sender=None,
    llmconfig=None,
    tools=None,
    datalogger=None,
    persist_path=None,
    tool_manager=None,
    return_template=True,
    **kwargs,
) -> ScoreTemplate | float:
    """
    Scores a given context using a language model, with the option to score multiple instances.

    Args:
        sentence (str | list | dict): The given context to score.
        num_instances (int): The number of instances to score (default: 1).
        instruction (Optional[str]): The instruction for scoring the context.
        score_range (tuple): The range of valid scores (default: (1, 10)).
        inclusive (bool): Whether to include the endpoints of the score range (default: True).
        num_digit (int): The number of digits after the decimal point for the score (default: 0).
        confidence_score (bool): Whether to include the confidence score in the output (default: False).
        reason (bool): Whether to include the reason for the score in the output (default: False).
        retries (int): The number of retries for the API call (default: 2).
        delay (float): The initial delay between retries in seconds (default: 0.5).
        backoff_factor (float): The backoff factor for exponential delay between retries (default: 2).
        default_value (Optional[Any]): The default value to return if the API call fails (default: None).
        timeout (Optional[float]): The timeout for the API call in seconds (default: None).
        branch_name (Optional[str]): The name of the branch to use for scoring.
        system (Optional[Any]): The system configuration for the branch.
        messages (Optional[Any]): The messages to initialize the branch with.
        service (Optional[Any]): The service to use for scoring.
        sender (Optional[str]): The sender of the scoring request.
        llmconfig (Optional[Any]): The configuration for the language model.
        tools (Optional[Any]): The tools to use for scoring.
        datalogger (Optional[Any]): The data logger for the branch.
        persist_path (Optional[str]): The path to persist the branch data.
        tool_manager (Optional[Any]): The tool manager for the branch.
        return_template (bool): Whether to return the score template or only the score (default: True).
        **kwargs: Additional keyword arguments for the API call.

    Returns:
        ScoreTemplate | float: The score template with the scored context or the average score if `return_template` is False.
    """

    async def _inner(i=0):
        return await _score(
            sentence=sentence,
            instruction=instruction,
            score_range=score_range,
            inclusive=inclusive,
            num_digit=num_digit,
            confidence_score=confidence_score,
            reason=reason,
            retries=retries,
            delay=delay,
            backoff_factor=backoff_factor,
            default_value=default_value,
            timeout=timeout,
            branch_name=branch_name,
            system=system,
            messages=messages,
            service=service,
            sender=sender,
            llmconfig=llmconfig,
            tools=tools,
            datalogger=datalogger,
            persist_path=persist_path,
            tool_manager=tool_manager,
            **kwargs,
        )

    if num_instances == 1:
        _out = await _inner()
        return _out if return_template else _out.answer

    elif num_instances > 1:
        _outs = await func_call.alcall(range(num_instances), _inner)
        return _outs if return_template else np.mean([i.answer for i in _outs])



"""
This module contains the SelectTemplate class for selecting an item from given choices based on a given context.

The SelectTemplate class is a subclass of ScoredTemplate and provides functionality for selecting an item from
given choices using a language model. It includes fields for the input sentence, choices, selected answer,
confidence score, and reason for the selection.

The module also includes a function for selecting an item from given choices using the SelectTemplate class
and a language model.
"""

from enum import Enum
from pydantic import Field

from lionagi.libs import func_call, StringMatch
from ..prompt.prompt_template import ScoredTemplate
from ..branch import Branch


class SelectTemplate(ScoredTemplate):
    """
    A class for selecting an item from given choices based on a given context.

    Attributes:
        template_name (str): The name of the select template (default: "default_select").
        sentence (str | list | dict): The given context.
        answer (Enum | str): The selected item from the given choices.
        signature (str): The signature indicating the input and output fields (default: "sentence -> answer").

    Methods:
        __init__(self, sentence=None, choices=None, instruction=None, reason=False, confidence_score=False, **kwargs):
            Initializes a new instance of the SelectTemplate class.
    """

    template_name: str = "default_select"
    sentence: str | list | dict = Field(
        default_factory=str, description="the given context"
    )
    answer: Enum | str = Field(
        default_factory=str, description="selection from given choices"
    )

    signature: str = "sentence -> answer"

    def __init__(
        self,
        sentence=None,
        choices=None,
        instruction=None,
        reason=False,
        confidence_score=False,
        **kwargs,
    ):
        """
        Initializes a new instance of the SelectTemplate class.

        Args:
            sentence (Optional[str | list | dict]): The given context.
            choices (Optional[list]): The list of choices to select from.
            instruction (Optional[str]): The instruction for selection.
            reason (bool): Whether to include the reason for the selection in the output (default: False).
            confidence_score (bool): Whether to include the confidence score in the output (default: False).
            **kwargs: Additional keyword arguments.
        """
        super().__init__(**kwargs)

        self.sentence = sentence
        self.choices = choices
        self.task = f"select 1 item, from provided choices {choices}."
        if instruction:
            self.task += f"objetive {instruction}."

        if reason:
            self.output_fields.append("reason")

        if confidence_score:
            self.output_fields.append("confidence_score")


async def select(
    sentence,
    choices=None,
    instruction=None,
    confidence_score=False,
    reason=False,
    retries=2,
    delay=0.5,
    backoff_factor=2,
    default_value=None,
    timeout=None,
    branch_name=None,
    system=None,
    messages=None,
    service=None,
    sender=None,
    llmconfig=None,
    tools=None,
    datalogger=None,
    persist_path=None,
    tool_manager=None,
    **kwargs,
):
    """
    Selects an item from given choices based on a given context using a language model.

    Args:
        sentence (str | list | dict): The given context.
        choices (Optional[list]): The list of choices to select from.
        instruction (Optional[str]): The instruction for selection.
        confidence_score (bool): Whether to include the confidence score in the output (default: False).
        reason (bool): Whether to include the reason for the selection in the output (default: False).
        retries (int): The number of retries for the API call (default: 2).
        delay (float): The initial delay between retries in seconds (default: 0.5).
        backoff_factor (float): The backoff factor for exponential delay between retries (default: 2).
        default_value (Optional[Any]): The default value to return if the API call fails (default: None).
        timeout (Optional[float]): The timeout for the API call in seconds (default: None).
        branch_name (Optional[str]): The name of the branch to use for selection.
        system (Optional[Any]): The system configuration for the branch.
        messages (Optional[Any]): The messages to initialize the branch with.
        service (Optional[Any]): The service to use for selection.
        sender (Optional[str]): The sender of the selection request.
        llmconfig (Optional[Any]): The configuration for the language model.
        tools (Optional[Any]): The tools to use for selection.
        datalogger (Optional[Any]): The data logger for the branch.
        persist_path (Optional[str]): The path to persist the branch data.
        tool_manager (Optional[Any]): The tool manager for the branch.
        **kwargs: Additional keyword arguments for the API call.

    Returns:
        SelectTemplate: The select template with the selected item.
    """
    branch = Branch(
        name=branch_name,
        system=system,
        messages=messages,
        service=service,
        sender=sender,
        llmconfig=llmconfig,
        tools=tools,
        datalogger=datalogger,
        persist_path=persist_path,
        tool_manager=tool_manager,
    )

    _template = SelectTemplate(
        sentence=sentence,
        choices=choices,
        instruction=instruction,
        confidence_score=confidence_score,
        reason=reason,
    )

    await func_call.rcall(
        branch.chat,
        prompt_template=_template,
        retries=retries,
        delay=delay,
        backoff_factor=backoff_factor,
        default=default_value,
        timeout=timeout,
        **kwargs,
    )

    ans = _template.answer

    if ans not in _template.choices:
        _template.answer = StringMatch.choose_most_similar(ans, _template.choices)

    return _template



# from lionagi.core.branch.branch import Branch
# from lionagi.core.session.session import Session

# import unittest
# from unittest.mock import patch, call, MagicMock
# import pandas as pd
# import json
# from datetime import datetime


# class TestSession(unittest.TestCase):

# 	def setUp(self):
# 		mock_branch = MagicMock()

# 		mock_branch.to_csv_file = MagicMock()
# 		mock_branch.to_json_file = MagicMock()

# 		mock_datalogger = MagicMock()
# 		mock_datalogger.to_csv_file = MagicMock()
# 		mock_datalogger.to_json_file = MagicMock()

# 		# Assign the mocked datalogger to the mock_branch
# 		mock_branch.datalogger = mock_datalogger

# 		self.branch1 = mock_branch(name="branch1")
# 		self.branch1.messages = pd.DataFrame(
# 			[{
# 				"node_id": "1", "timestamp": "2021-01-01 00:00:00",
# 				"role": "system", "sender": "system",
# 				"content": json.dumps({"system_info": "System message"}),
# 			}]
# 		)
# 		self.branch2 = mock_branch(name="branch2")
# 		self.branch1.messages = pd.DataFrame(
# 			[{
# 				"node_id": "2", "timestamp": "2021-01-01 00:01:00",
# 				"role": "user", "sender": "user1",
# 				"content": json.dumps({"instruction": "User message"}),
# 			}]
# 		)

# 		branches = {"branch1": self.branch1, "branch2": self.branch2}

# 		self.session = Session(
# 			branches=branches, default_branch_name="branch1",
# 			default_branch=branches["branch1"], )
# 		self.session.mail_manager = MagicMock()

# 	# def test_from_csv_initialization(self):
# 	#     """Test Session initialization from a CSV file."""
# 	#     mock_df = pd.DataFrame(
# 	#         {
# 	#             "node_id": ["1", "2"],
# 	#             "timestamp": [datetime(2021, 1, 1), datetime(2021, 1, 1)],
# 	#             "role": ["system", "user"],
# 	#             "sender": ["system", "user1"],
# 	#             "content": [
# 	#                 json.dumps({"system_info": "System message"}),
# 	#                 json.dumps({"instruction": "User message"}),
# 	#             ],
# 	#         }
# 	#     )
# 	#     filepath = "path/to/mock.csv"

# 	#     with patch("pandas.read_csv", return_value=mock_df) as mock_read_csv:
# 	#         session = Session.from_csv(filepath)
# 	#         mock_read_csv.assert_called_once_with(filepath)
# 	#         pd.testing.assert_frame_equal(session.messages, mock_df)

# 	# def test_from_json_initialization(self):
# 	#     """Test Session initialization from a CSV file."""
# 	#     mock_df = pd.DataFrame(
# 	#         {
# 	#             "node_id": ["1", "2"],
# 	#             "timestamp": [datetime(2021, 1, 1), datetime(2021, 1, 1)],
# 	#             "role": ["system", "user"],
# 	#             "sender": ["system", "user1"],
# 	#             "content": [
# 	#                 json.dumps({"system_info": "System message"}),
# 	#                 json.dumps({"instruction": "User message"}),
# 	#             ],
# 	#         }
# 	#     )
# 	#     filepath = "path/to/mock.json"

# 	#     with patch("pandas.read_json", return_value=mock_df) as mock_read_json:
# 	#         session = Session.from_json(filepath)
# 	#         mock_read_json.assert_called_once_with(filepath)
# 	#         pd.testing.assert_frame_equal(session.messages, mock_df)

# 	def test_to_csv_file(self):
# 		"""Ensure to_csv_file calls each branch's to_csv_file method."""
# 		filename = "test_export.csv"
# 		self.session.to_csv_file(filename=filename)

# 		for name, branch in self.session.branches.items():
# 			# Verify it was called twice
# 			self.assertEqual(branch.to_csv_file.call_count, 2)

# 			# Verify the arguments of the last call
# 			expected_filename = f"{name}_{filename}"
# 			self.assertIn(
# 				expected_filename,
# 				["branch1_test_export.csv", "branch2_test_export.csv"], )

# 	def test_to_json_file(self):
# 		"""Ensure to_json_file calls each branch's to_json_file method."""
# 		filename = "test_export.json"
# 		self.session.to_json_file(filename=filename)

# 		for name, branch in self.session.branches.items():
# 			# Verify it was called twice
# 			self.assertEqual(branch.to_json_file.call_count, 2)

# 			# Verify the arguments of the last call
# 			expected_filename = f"{name}_{filename}"
# 			self.assertIn(
# 				expected_filename,
# 				["branch1_test_export.json", "branch2_test_export.json"], )

# 	def test_log_to_csv(self):
# 		"""Ensure log_to_csv calls each branch's log_to_csv method."""
# 		filename = "test_export.csv"
# 		self.session.log_to_csv(filename=filename)

# 		for name, branch in self.session.branches.items():
# 			# Verify it was called twice
# 			self.assertEqual(branch.log_to_csv.call_count, 2)

# 			# Verify the arguments of the last call
# 			expected_filename = f"{name}_{filename}"
# 			self.assertIn(
# 				expected_filename,
# 				["branch1_test_export.csv", "branch2_test_export.csv"], )

# 	def test_log_to_json(self):
# 		"""Ensure log_to_json calls each branch's log_to_json method."""
# 		filename = "test_export.json"
# 		self.session.log_to_json(filename=filename)

# 		for name, branch in self.session.branches.items():
# 			# Verify it was called twice
# 			self.assertEqual(branch.log_to_json.call_count, 2)

# 			# Verify the arguments of the last call
# 			expected_filename = f"{name}_{filename}"
# 			self.assertIn(
# 				expected_filename,
# 				["branch1_test_export.json", "branch2_test_export.json"], )

# 	def test_all_messages(self):
# 		"""Test aggregation of all messages across branches."""
# 		expected_df = pd.concat(
# 			[self.branch1.messages, self.branch2.messages], ignore_index=True
# 		)

# 		actual_df = self.session.all_messages

# 		pd.testing.assert_frame_equal(actual_df, expected_df)

# 	def test_new_branch_creation(self):
# 		"""Test creating a new branch successfully."""
# 		branch_name = "test_branch"
# 		self.session.new_branch(branch_name=branch_name)
# 		self.assertIn(branch_name, self.session.branches)

# 	def test_new_branch_duplicate_name(self):
# 		"""Test error handling for duplicate branch names."""
# 		branch_name = "test_branch"
# 		self.session.new_branch(branch_name=branch_name)
# 		with self.assertRaises(ValueError):
# 			self.session.new_branch(branch_name=branch_name)

# 	def test_get_branch_by_name(self):
# 		"""Test retrieving a branch by its name."""
# 		branch_name = "test_branch"
# 		self.session.new_branch(branch_name=branch_name)
# 		branch = self.session.get_branch(branch_name)
# 		self.assertIsInstance(branch, Branch)

# 	def test_get_branch_invalid_name(self):
# 		"""Test error handling for invalid branch names."""
# 		with self.assertRaises(ValueError):
# 			self.session.get_branch("nonexistent_branch")

# 	def test_change_default_branch(self):
# 		"""Test changing the default branch."""
# 		branch_name = "new_default"
# 		self.session.new_branch(branch_name=branch_name)
# 		self.session.change_default_branch(branch_name)
# 		self.assertEqual(self.session.default_branch_name, branch_name)

# 	def test_delete_branch(self):
# 		"""Test deleting a branch."""
# 		branch_name = "test_branch"
# 		self.session.new_branch(branch_name=branch_name)
# 		self.session.delete_branch(branch_name)
# 		self.assertNotIn(branch_name, self.session.branches)

# 	def test_delete_default_branch_error(self):
# 		"""Test error when trying to delete the default branch."""
# 		with self.assertRaises(ValueError):
# 			self.session.delete_branch(self.session.default_branch_name)

# 	def test_merge_branch(self):
# 		"""Test merging two branches."""
# 		from_branch = "source_branch"
# 		to_branch = "target_branch"
# 		self.session.new_branch(branch_name=from_branch)
# 		self.session.new_branch(branch_name=to_branch)
# 		self.session.merge_branch(
# 			from_=from_branch, to_branch=to_branch, del_=True
# 			)
# 		self.assertIn(to_branch, self.session.branches)
# 		self.assertNotIn(from_branch, self.session.branches)

# 	def test_collect_from_specified_branches(self):
# 		"""Test collecting requests from specified branches."""
# 		self.session.collect(from_=["branch1"])
# 		self.assertEqual(self.session.mail_manager.collect.call_count, 1)

# 	def test_collect_from_all_branches(self):
# 		"""Test collecting requests from all branches."""
# 		self.session.collect()
# 		self.assertEqual(self.session.mail_manager.collect.call_count, 2)

# 	def test_send_to_specified_branches(self):
# 		"""Test sending requests to specified branches."""
# 		self.session.send(to_=["branch_1"])
# 		self.assertEqual(self.session.mail_manager.send.call_count, 1)

# 	def test_send_to_all_branches(self):
# 		"""Test sending requests to all branches."""
# 		self.session.send()
# 		self.assertEqual(self.session.mail_manager.send.call_count, 2)

# 	def test_collect_send_all_without_receive_all(self):
# 		"""Test collecting and sending requests across all branches without receiving."""
# 		self.session.collect_send_all()
# 		self.assertEqual(self.session.mail_manager.collect.call_count, 2)
# 		self.assertEqual(self.session.mail_manager.send.call_count, 2)
# 		self.branch1.receive_all.assert_not_called()
# 		self.branch2.receive_all.assert_not_called()

# 	def test_collect_send_all_with_receive_all(self):
# 		"""Test collecting and sending requests across all branches with receiving."""
# 		self.session.collect_send_all(receive_all=True)
# 		self.branch1.receive_all.assert_called()
# 		self.branch2.receive_all.assert_called()


# if __name__ == "__main__":
# 	unittest.main()



from lionagi.libs.sys_util import SysUtil
import lionagi.libs.ln_convert as convert
from lionagi.libs.ln_api import BaseService
from lionagi.integrations.config.mlx_configs import model


class MlXService(BaseService):
    def __init__(self, model=model, **kwargs):

        SysUtil.check_import("mlx_lm")

        from mlx_lm import load, generate

        super().__init__()

        model_, tokenizer = load(model, **kwargs)

        self.model_name = model
        self.model = model_
        self.tokenizer = tokenizer
        self.generate = generate

    async def serve_chat(self, messages, **kwargs):
        if "verbose" not in kwargs.keys():
            verbose = True

        prompts = [
            convert.to_dict(msg["content"])["instruction"]
            for msg in messages
            if msg["role"] == "user"
        ]

        payload = {"messages": messages}

        try:
            response = self.generate(
                self.model,
                self.tokenizer,
                prompt=f"{prompts[-1]} \nOutput: ",
                verbose=verbose,
            )
            completion = {"model": self.model_name, "choices": [{"message": response}]}

            return payload, completion
        except Exception as e:
            self.status_tracker.num_tasks_failed += 1
            raise e
